package CPU;

/**
 * Object derived from CPU, used only when emulating QL screen and copying from there to display.
 * 
 * @author and copyright (C) Wolfgang Lenerz 2016-2017.
 * @version 
 * 1.00 derived from MC68000Cpu v.2.10.
 */
public class CPUforScreenEmulation extends MC68000Cpu
{
    private boolean copyQLScreen=false;
    
    /**
     * Creates the object.
     * 
     * @param size the size of ram in bytes.
     * @param screen the screen to use.
     * @param iniFile my .ini file with options to set.
     * @param romSize the size of the ROM image. Loaded after the RAM image.
     */
    public CPUforScreenEmulation(int size,screens.Screen screen,inifile.IniFile iniFile,int romSize)
    {
        super(size,screen,iniFile,romSize);
    }
    
    /**
     * Creates the object.
     * 
     * @param size the size of ram in bytes.
     * @param iniFile my .ini file with options to set.
     * @param romSize the size of the ROM image. Loaded after the RAM image.
     * @param videoRamSize size,in bytes, of screen memory to be allocated here.
     */
       public CPUforScreenEmulation(int size,inifile.IniFile iniFile,int romSize,int videoRamSize)
    {
        super(size,iniFile,romSize,videoRamSize);
    }
        
    /**
     * Writes a byte to memory.
     * 
     * @param address where to write to.
     * 
     * @param value the byte to write (in the LSB of the int).
     */
    @Override
    public final void writeMemoryByte(int address, int value)
    {  
        address&=MC68000Cpu.CUT_OFF;                             // !!! remove higher bits  also means that address won't be <0
        if (address>this.totMemSize)
            return;    
        int addr=address/2;
        int val=value;
        char res=(this.mainMemory[addr]);
        if((address&1)!=0)
        {
            res&=0xff00;
            value&=0xff;
        }
        else
        {
            res&=0x00ff;
            value=(value<<8)&0xff00;
        }
        value|=res;
        this.mainMemory[addr]=(char)value;                              // ,xxxxxxxxxxxxxxxxxx
        if(address>= this.screenStart)
        {
            this.screen.writeByteToScreen(address,val,value);       // trying to write to screen?
        }
        if (this.copyQLScreen && address>=0x20000 && address< 0x28000)
            this.screen.copyScreen(this, addr*2,value);
    }
    
    /**
     * Writes a word to memory.
     * @param address where to write to.
     * @param value the word to write (in the LSW of the int).
     */
    @Override
    public final void writeMemoryWord(int address, int value)
    {   
        // debugMem(address);
        address&=MC68000Cpu.CUT_OFF;                            // !!! remove higher bits  also means that address won't be <0
        if (address>this.totRamSizeForWord)
            return;  
        this.mainMemory[address/2]= (char)(value);
        if(address>= this.screenStart)
        {
            this.screen.writeWordToScreen(address,value);   // trying to write screen?     
        }
        if (this.copyQLScreen && address>=0x20000 && address< 0x28000)
            this.screen.copyScreen(this, address,value);
    }
    
    /**
     * Writes a short to memory.
     * 
     * @param address where to write to.
     * @param value the short to write.
     */
    @Override
    public final void writeMemoryShort(int address, short value) //       xxxxxxxxxxxxxxxxxx
    {   
        // debugMem(address);
        address&=MC68000Cpu.CUT_OFF;                            // !!! remove higher bits  also means that address won't be <0
        if (address>this.totRamSizeForWord)
            return;  
        this.mainMemory[address/2]= (char) value;
        if(address>= this.screenStart)
        {
            this.screen.writeWordToScreen(address,value);   // trying to write screen?
        }
        if (this.copyQLScreen && address>=0x20000 && address< 0x28000)
            this.screen.copyScreen(this, address,value);
    }
    
    /**
     * Writes a long to the memory.
     * @param address the address where to write to.
     * @param value the value to write.
     */
    @Override
    public final void writeMemoryLong(int address, int value)
    {  
        // debugMem(address);
        address&=MC68000Cpu.CUT_OFF;                            // !!! remove higher bits  also means that address won't be <0
        if (address>this.totRamSizeForLong)
            return;  
        int addr=address/2;
        this.mainMemory[addr]= (char)(value>>>16); 
        this.mainMemory[addr+1]= (char)(value); 
        
        if(address>= this.screenStart)
        {
            this.screen.writeLongToScreen(address,value);   // trying to write screen?
        }
        if (this.copyQLScreen && address>=0x20000 && address< 0x27ffe)
        {
            this.screen.copyScreen(this, address,value>>>16);
            this.screen.copyScreen(this, address+2,value&0xffff);
        }
    }
     
    /**
     * This switches QL screen emulation on or off. If on, any writes to the QL screen area are copied to the true display.
     * 
     * @param origins origins in y screen which should correspond to (0,0) in QL screen.
     * @param QLScreenMode the screen mode the QL screen is supposed to be in.
     */
    @Override
    public void setCopyScreen(int QLScreenMode,int origins)
    {
        this.copyQLScreen=QLScreenMode!=0;
        this.screen.setQLEmulationMode (QLScreenMode);
        this.data_regs[0]=0;
        this.reg_sr|=4;
    }
    
    /**
     * Reads bytes from a ByteBuffer and writes them into memory. Possibly copies QL screen memory to display.
     * 
     * @param address where to read to.
     * @param nbrOfBytes how many bytes are to be read.
     * @param buffer the buffer to read from.
     * @param startInBuffer where in the buffer to start reading from.
     * @param specialRead =<code>true</code> ONLY when loading the OS, must be <code>false</code> at all other times!
     * 
     * @return nbr of bytes read.
     */
    @Override
    protected int readFromBuffer(int address,int nbrOfBytes,java.nio.ByteBuffer buffer,int startInBuffer,boolean specialRead)
    {
        super.readFromBuffer(address, nbrOfBytes, buffer, startInBuffer, specialRead);
       if (this.copyQLScreen && !specialRead && (address>=0x20000 && address+nbrOfBytes<0x28000)) 
        {
            this.screen.setDisplayRegion(address,nbrOfBytes, this.mainMemory);
        }
        return nbrOfBytes;
    }
}
