
package CPU.instructions.LSR;

/**
 * The LSR instruction where data to be shifted is in a data reg and the shift count is immediate.
 * 1110ccc0ss001rrr = 0xe008   
 * 
 * mmm = mode 010
 * rrr = data register to be shifted
 * ccc = the shift count
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * 
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2014 - 2020.
 * @version
 *  1.02 speedup by revamping
 *  1.01 LSR.L corrected, used >> instead of >>>
 * Based on code by Tony Headford, see his licence in accompanying file.
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class LSRimm implements InstructionSet
{
    private static final int B_MASK = 0x80;
    
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                       // byte, word and long only
        {
            switch (sz)
            {
                case 0:                                     // shift lsb
                    base = 0xe008;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrShifts = ((opcode>>9)&0x7);      // how often we shift
                            cpu.reg_sr&=0xffe0;                     // all status flags set to 0 (V is always cleared here)
                            boolean lastBitSet;                     // will be true if the value of the last bit to be shifted out = 1
                            if (nbrShifts==0)                       // 0 means shift by 8
                            {
                                lastBitSet= (cpu.data_regs[reg]&B_MASK)!=0;
                                cpu.data_regs[reg]&=0xffffff00;     // shift by 8 = resulting byte is 0
                                cpu.reg_sr+=4;                      // set Z flag 
                            }
                            else
                            {
                                int val=cpu.data_regs[reg]&0xff;    // byte to be shifted
                                lastBitSet= (cpu.data_regs[reg]&(1<<(nbrShifts-1)))!=0;
                                val>>=nbrShifts;
                                if (val==0)
                                    cpu.reg_sr+=4;                  // set Z flag 
                                cpu.data_regs[reg]&=0xffffff00;     //get rid of ls byte
                                cpu.data_regs[reg]|=val;            // and fill it in with new value
                            }
                            if (lastBitSet)
                                cpu.reg_sr|=17;                     // set X + C flags   NB N flag can never be set here, result is never negative!
                            
                            /* remove at next review
                            int nbrShifts = ((opcode>>9)&0x7 );     // how often we shift
                            if (nbrShifts==0)
                                nbrShifts=8;
                            int val=cpu.data_regs[opcode&0x7]&0xff;// byte to be shifted
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&1;                      // lowest bit before shift
                                val>>>=1;                           // shift one bit right
                            }
                            val&=0xff;
                            cpu.data_regs[opcode&0x7]&=0xffffff00; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xffe0;                // all status flags set to 0 (V is always cleared here)
                            if ((val&0x80)!=0)                      // ms bit set
                                cpu.reg_sr+=8;                 // so set N flag
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (lastOut!=0)
                                cpu.reg_sr|=17;                // set X + C flags
                            */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };  
                    break;
          
                case 1:                                             // word sized shift
                    base = 0xe048;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrShifts = ((opcode>>9)&0x7);      // how often we shift
                            if (nbrShifts==0)                       // 0 means shift by 8
                                nbrShifts = 8;
                            int val=cpu.data_regs[reg]&0xffff;      // word to be shifted
                            boolean lastBitSet= (cpu.data_regs[reg]&(1<<(nbrShifts-1)))!=0; // will be true if the value of the last bit to be shifted out = 1
                            val>>=nbrShifts;
                            cpu.data_regs[reg]&=0xffff0000;         //get rid of low word
                            cpu.data_regs[reg]|=val;                // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                     // all status flags set to 0 (V is always cleared here)
                            if (val==0)
                                cpu.reg_sr+=4;                      // set Z flag
                            if (lastBitSet)
                                cpu.reg_sr|=17;                     // set X + C flags   NB N flag can never be set here, result is never negative!
                            
                            /* remove at next review
                            int nbrShifts = ((opcode>>9)&0x7 );     // how often we shift
                            if (nbrShifts==0)
                            nbrShifts=8;
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// word to be shifted
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                            lastOut=val&1;                      // highest bit before shift
                            val>>>=1;                           // shift one bit right
                            }
                            val&=0xffff;
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xffe0;                // all status flags set to 0
                            if ((val&0x8000)!=0)                    // ms bit set
                            cpu.reg_sr|=8;                 // so set N flag
                            else if (val==0)
                            cpu.reg_sr|=4;                 // set Z flag
                            if (lastOut!=0)
                            cpu.reg_sr|=17;                // set X + C flags */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                    };  
                    break;
                    
                default:
                    base = 0xe088;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrShifts = ((opcode>>9)&0x7);      // how often we shift
                            if (nbrShifts==0)                       // 0 means shift by 8
                                nbrShifts = 8;
                            boolean lastBitSet= (cpu.data_regs[reg]&(1<<(nbrShifts-1)))!=0; // will be true if the value of the last bit to be shifted out = 1
                            cpu.data_regs[reg]>>>=nbrShifts;         
                            cpu.reg_sr&=0xffe0;                     // all status flags set to 0 (V is always cleared here)
                            if (cpu.data_regs[reg]==0)
                                cpu.reg_sr+=4;                      // set Z flag 
                            if (lastBitSet)
                                cpu.reg_sr|=17;                     // set X + C flags   NB N flag can never be set here, result is never negative!
                        
                            /* remove at next review
                            int nbrShifts = ((opcode>>9)&0x7 );     // how often we shift
                            if (nbrShifts==0)
                                nbrShifts=8;
                            int val=cpu.data_regs[opcode&0x7]; // long word to be shifted
                            int lastOut=0;                          // current state of highest bit
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&1;
                                val>>>=1;                           // shift one bit right
                            }
                            cpu.data_regs[opcode&0x7]=val;     // set new value
                            cpu.reg_sr&=0xffe0;                // all status flags set to 0
                            if (val<0)                // ms bit set (could be if 0 shift count)
                                cpu.reg_sr+=8;                 // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (lastOut!=0)
                                cpu.reg_sr|=17;                // set X + C flags
                        */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
        int count = (opcode >> 9) & 0x07;
        if(count == 0)
                count = 8;
        DisassembledOperand src = new DisassembledOperand("#" + count);
	DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
        return new DisassembledInstruction(address, opcode, "lsr" + sz.ext(), src, dst);
    }
}
