package CPU.instructions.ROX;

/**
 * The ROXR instruction where data to be rotated is in a data reg and the rotate count is in another data reg.
 * 1110ccc0ss110rrr = 0xe030   
 * 
 * mmm = mode 010
 * rrr = data register to be shifted
 * ccc = dataregister containing the shift count (modulo 64)
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * 
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2014.
 * 
 * Based on code by Tony Headford, see his licence in accompanying file
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class ROXRreg implements InstructionSet
{
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                           // byte, word and long only
        {
            switch (sz)
            {
                case 0:                                         // rotate byte
                    base = 0xe030;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrRotats = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[opcode&0x7]&0xff;// byte to be shifted
                            int lastOut=0;
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            for (int cnt=0;cnt<nbrRotats;cnt++)
                            {
                                lastOut=val&0x1;                // lowest bit before shift
                                val>>>=1;                       // shift one bit down
                                if (xBit!=0)
                                    val|=0x80;
                                xBit=lastOut;
                            }
                            cpu.data_regs[opcode&0x7]&=0xffffff00; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;     // and fill it in with new value
                            cpu.reg_sr&=0xfff0;                 // all status flags set to 0, except for X
                            if ((val&0x80)!=0)                  // ms bit set
                                cpu.reg_sr+=8;                  // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                  // set Z flag
                            if (nbrRotats!=0)                   // X flag is unaffected by 0 count shift, C is cleared by it
                            {
                                if (lastOut!=0)
                                    cpu.reg_sr|=17;             // set X + C flags
                                else
                                    cpu.reg_sr&=~16;            // clear X flag (C flag already cleared)
                            }
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };  
                    break;
                
                case 1:
                    base = 0xe070;                              // word sized rotate
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrRotats = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// word to be shifted
                            int lastOut=0;
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            for (int cnt=0;cnt<nbrRotats;cnt++)
                            {
                                lastOut=val&0x1;                // lowest bit before shift
                                val>>>=1;                       // shift one bit down
                                if (xBit!=0)
                                    val|=0x8000;
                                xBit=lastOut;
                            }
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;     // and fill it in with new value
                            cpu.reg_sr&=0xfff0;                 // all status flags set to 0, except for X
                            if ((val&0x8000)!=0)                // ms bit set
                                cpu.reg_sr+=8;                  // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                  // set Z flag
                            if (nbrRotats!=0)                   // X flag is unaffected by 0 count shift, C is cleared by it
                            {
                                if (lastOut!=0)
                                    cpu.reg_sr|=17;             // set X + C flags
                                else
                                    cpu.reg_sr&=~16;            // clear X flag (C flag already cleared)
                            }
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                        
                    }; 
                    break;
                    
                default:                                        // rotate LW
                    base = 0xe0b0;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrRotats = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often we shift
                            int val=cpu.data_regs[opcode&0x7]; // long word to be shifted
                            int lastOut=0;
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            for (int cnt=0;cnt<nbrRotats;cnt++)
                            {
                                lastOut=val&0x1;                // lowest bit before shift
                                val>>>=1;                       // shift one bit down
                                if (xBit!=0)
                                    val|=0x80000000;
                                xBit=lastOut;
                            }
                            cpu.data_regs[opcode&0x7]=val;     
                            cpu.reg_sr&=0xfff0;                 // all status flags set to 0, except for X
                            if (val<0)                          // ms bit set
                                cpu.reg_sr+=8;                  // so set N flag if (val==0)
                            else if (val==0)
                                cpu.reg_sr+=4;                  // set Z flag
                            if (nbrRotats!=0)                   // X flag is unaffected by 0 count shift, C is cleared by it
                            {   
                                if (lastOut!=0)
                                    cpu.reg_sr|=17;             // set X + C flags
                                else
                                    cpu.reg_sr&=~16;            // clear X flag (C flag already cleared)
                            }
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
           //reg mode
            DisassembledOperand src = new DisassembledOperand("d" + ((opcode >> 9) & 0x07));
            DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
            return new DisassembledInstruction(address, opcode, "roxr" + sz.ext(), src, dst);
    }
}
