package drivers;

import smsqmulator.IPHandler;
import smsqmulator.Types;

/**
 * This handles UDP connections.
 * This is more complicated than Marcel's C code, since Java has three different objects for socket:
 * Socket (client side TCP socket), ServerSocket and DatagramSocket (UDP),and they are not interchangeable.
 * 
 * This class is not thread safe.
 * Only 50 connections can be open at once.
 * 
 * @author and copyright (c) wolfgang lenerz 2015-2020
 * 
 * @version 
 * 1.04 contains the error strings, class IPError is abandoned, sockets are in an array not a map ( a map with a small integer as key, what was I tinking?).
 *      writeCStringUnsafe used, better, nurt unsafe handling of  gethostbyname I/O op.
 * 1.03 IOB.FLIN, make D1, D2 comparison on (unsigned) word only, not LW.
 * 1.02 handleTrap added case 7.
 * 1.01 interface change for get_netname (case 6), needs at least 3.31.
 * 1.00 implements only SCK and TCP, trap3, 1-7,50,51,53,58,59,5b,5e,62,7b,7c. No UDP.
 * 0.00 initial version, just with trap#3;1-7.
 * 
 */
public class UDPHandler
{
    private final UDPSocket[]   sockets;
    private static final int    MAX_SOCKETS=50;
    private static final int    MAX_PACKET_SIZE = 1470;        // max size of payload for one udp packet
    
    /**
     * These are the I/O codes I'm supposed to handle.
     */
    private static final int IOB_TEST   = 0x00;
    private static final int IOB_FBYT   = 0x01;
    private static final int IOB_FLIN   = 0x02;
    private static final int IOB_FMUL   = 0x03;
    private static final int IOB_SBYT   = 0x05;    
    private static final int IOB_SMUL   = 0x07;
    private static final int IOF_LOAD   = 0x48;
    private static final int IOF_SAVE   = 0x49;

//    private static final int IP_LISTEN  = 0x50;
    private static final int IP_SEND    = 0x51;
    private static final int IP_SENDTO  = 0x52;
    private static final int IP_RECV    = 0x53;
    private static final int IP_RECVFM  = 0x54;
//    private static final int IP_GETOPT  = 0x55;
  //  private static final int IP_SETOPT  = 0x56;
 //   private static final int IP_SHUTDWN = 0x57;
    private static final int IP_BIND    = 0x58;
    private static final int IP_CONNECT = 0x59;
    private static final int IP_GETHOSTNAME = 0x5b;
  //  private static final int IP_GETSOCKNAME = 0x5c;
    private static final int IP_GETPEERNAME = 0x5d;
    private static final int IP_GETHOSTBYNAME= 0x5e;
  //  private static final int IP_GETHOSTBYADDR= 0x5f;
//    private static final int IP_SETHOSTENT       0x60
//    private static final int IP_ENDHOSTENT       0x61
    private static final int IP_H_ERRNO     = 0x62;
 /*
    private static final int IP_GETSERVBYNAME   = 0x64;
    private static final int IP_GETSERVBYPORT   = 0x65;
    private static final int IP_GETPROTOBYNAME  = 0x6e;
    private static final int IP_GETPROTOBYNUMBER= 0x6f;
    private static final int IP_INET_ATON   = 0x72;
    private static final int IP_INET_ADDR   = 0x73;
    private static final int IP_INET_NETWORK= 0x74;
    private static final int IP_INET_NTOA   = 0x75;
    private static final int IP_IOCTL       = 0x79;
    private static final int IP_GETDOMAIN   = 0x7a;
*/
    private static final int IP_H_STRERROR  = 0x7b;
    private static final int IP_ERRNO       = 0x7c;
    private static final int IP_DISCONNECT  = 0x7f;
   
    /**
     * Error strings, always in english.
     */ 
    public static final String[] IP_ERROR = 
    {
        "NO ERROR",
        "UNTYPED SOCKET - can't do I/O",
        "GENERIC SOCKET - can't do I/O",
        "SOCKET IS ALREADY CONNECTED - can't connect again",
        "WRONG HOST ADDRESS :- The host address is wrong or malformed",
        "WRONG PORT NUMBER",
        "I/O EXCEPTION DURING SOCKET CREATION",
        "SECURITY MANAGER VIOLATION DURING SOCKET CREATION",
        "UNKNOWN ERROR",
        "SOCKET IS NOT CONNECTED",
        "I/O EXCEPTION WHEN SENDING BYTES",
        "I/O EXCEPTION WHEN RECEIVING BYTES",
        "ILLEGAL SOCKET CONVERSION",
        "SERVER CHANNEL CLOSED",
        "FUNCTION NOT IMPLEMENTED",
        "CONNECTION REFUSED",
        "SOCKET ALREADY BOUND",
        "WRONG OR ILLEGAL SOURCE IP ADDRESS", //17
    };
    
    private int lastError;
    
    /**
     * Creates the object.
     */
    public UDPHandler()
    {
        this.sockets=new UDPSocket[MAX_SOCKETS];
    }
    
    /**
     * This is a very primitive way to find a free entry in the array.
     * @return a new free entry for a socket in the array.
     */
    private int getNewKey()
    {
        for (int i=1;i<MAX_SOCKETS;i++)
        {
            if (this.sockets[i]==null)
                return i;
        }
        return -1;
    }
    
    /**
     * This handles all the SMSQE "trap" calls related to this device.
     * 
     * @param cpu  the cpu used.
     */
    public void handleTrap(CPU.MC68000Cpu cpu)
    {
        int d0;
        UDPSocket socket=null;
                
        switch (cpu.data_regs[0])                               // what operation should we do?
        {
            case 1:                                             // open a socket or make an accept call
                socket=new UDPSocket(cpu);                      // try to open a socket (this changes cpu.data_regs[0]!)
                if (cpu.data_regs[0]==0)                        
                {    
                    int key = getNewKey();                      // new socket opened OK
                    if (key==-1)
                    {
                        cpu.data_regs[0]=Types.ERR_IMEM;
                    }
                    else
                    {
                        this.sockets[key]= socket;              // accept call
                        cpu.data_regs[7]=-key;                  // put negative key into D7
                    }
                }
                break;
                
            case 2:                                             // do I/O
                d0=cpu.readMemoryLong(cpu.addr_regs[7]);        // get IO key
                cpu.addr_regs[7]+=4;                            // adjust stack
                socket=this.sockets[-cpu.addr_regs[0]];         // get the socket that's being addressed
                if (socket==null)
                    cpu.data_regs[0]=Types.ERR_ICHN;            // ooops
                else
                    handleIO (d0,cpu,socket);
                break;
                
            case 3:                                             // close socket
                socket=this.sockets[-cpu.addr_regs[0]];
                if (socket!=null)
                {
                    socket.close();                             // close it
                    this.sockets[-cpu.addr_regs[0]]=null;       // remove it from list
                }
                cpu.data_regs[0]=0;                             // show no error
                break;
                
            case 4:                                             // close all sockets, used at startup/reset
                for (int i =0;i<this.sockets.length;i++)
                {
                    if (this.sockets[i]!=null)
                    {
                        this.sockets[i].close();
                        this.sockets[i]=null;
                    }
                }
                cpu.data_regs[0]=0;
                break;
                
            case 5:                                             // get channel name into (a1), d2 = max length
                socket=this.sockets[-cpu.addr_regs[0]];
                String v;
                if (socket==null)
                    v="UNKNOWN SOCKET";
                else
                {
                    v=socket.getName();
                }
                
                int bytes=cpu.data_regs[2]&0xffff;
                smsqmulator.SmsqeString.writeSmsqeString(cpu.addr_regs[1],v,true,bytes,cpu);//write as many bytes as possible
                if (v.length()> bytes)
                    cpu.data_regs[0]=Types.ERR_IPAR;            // but signal buffer too small          
                else
                    cpu.data_regs[0]=0;                         // or signal all OK
                break;
                
            case 6:
                String s=getHName();                            // name of machine
                if (s==null)
                    s="Unknown name!";
                int l=s.length();
                int bufflen=cpu.readMemoryWord(cpu.addr_regs[1]+2);
                if (l>bufflen)
                {
                    cpu.data_regs[0]=Types.ERR_BFFL;
                }
                else
                {   
                    smsqmulator.SmsqeString.writeSmsqeString(cpu.readMemoryLong(cpu.addr_regs[1]+4), s, true, l,cpu);//write string
                    cpu.data_regs[0]=0;
                }
                break;
                
            case 7:
                try
                {
                    s=getLocalAddress();
                }
                catch (Exception u)
                { 
                    s="";
                }
                l=s.length();
                bufflen=cpu.readMemoryWord(cpu.addr_regs[1]+2);
                if (l>bufflen)
                {
                    cpu.data_regs[0]=Types.ERR_BFFL;
                }
                else
                {   
                    smsqmulator.SmsqeString.writeSmsqeString(cpu.readMemoryLong(cpu.addr_regs[1]+4), s, true, l,cpu);//write string
                    cpu.data_regs[0]=0;
                }    
                break;
            }
        
        if (cpu.data_regs[0]==0)
        {
            cpu.reg_sr|=4;
            this.lastError=0;
        }    
        else if (cpu.data_regs[0]<0)
        {
            cpu.reg_sr&=~4;                                     // set status reg according to D0;
            if (socket!=null)
                this.lastError=socket.getLastError();
        }
    }
    
    /**
     * This handles I/O.
     * 
     * The only real IO operations (where bytes are actually received/sent) allowed are : 
     *      test pending input, get 1 char, get 1 line ending in LF, get multiple bytes, scatter load
     *      send 1 byte, send multiple bytes, scatter save
     * D2 contains nbr of bytes or buffer remember for scatter load/save D2 is long, for other ops, D2 = word)
     * 
     * @param what type of operation
     * @param cpu the cpu used
     * @param m the socket this operation is to use
     */
    private void handleIO(int what,CPU.MC68000Cpu cpu,UDPSocket socket)
    {
        int bytes;
        cpu.data_regs[0]=0;             
        // preset no error
        int A1 = cpu.addr_regs[1];
        int D1 = cpu.data_regs[1];
        int D2 = cpu.data_regs[2];
        int result=0;

        switch (what)
        {
           case IOB_TEST:                                       // test pending input
                if (!socket.isPending())
                {
                    if (socket.isUDD() && socket.isClosed())
                        cpu.data_regs[0]=Types.ERR_EOF;
                    else
                        cpu.data_regs[0]=Types.ERR_NC;              // nothing is pending.
                }
                break;
                
            case IOB_FBYT:                                      // get ONE char    
                byte[]byt=new byte[1];
                result= socket.receive1Byte(cpu,byt);
                if (result>0)
                {
                    bytes=byt[0]&0xff;
                    cpu.data_regs[1]=(D1&0xffffff00)|bytes;// return the byte in D1
                    cpu.data_regs[0]=0;
                }
                else
                    cpu.data_regs[0]=result;
                break;
                
            case IOB_FLIN:                                      // get multiple bytes ending by LF into buffer at (a1), d2.w = buffer size
                int bytesGot;
                if (cpu.data_regs[3]==0)
                    bytesGot = 0;
                else
                    bytesGot=D1&0xffff;                         // nbr of bytes already got (if not first call)
                int totalBytesToGet = D2&0xffff;                // total nbr of bytes to get
                if (bytesGot>=totalBytesToGet)               
                {
                    cpu.data_regs[0]=Types.ERR_BFFL;            // we got all there was to get, but no LF found
                    break;                                      
                }
                int toGet = totalBytesToGet -bytesGot;          // I still need this many
                if (toGet> UDPHandler.MAX_PACKET_SIZE)
                    toGet = UDPHandler.MAX_PACKET_SIZE;
                
           //     byte[] mbytes = new byte[toGet];
                bytes= socket.receiveAndCheckForLF(cpu,toGet);  // get this many bytes, check for LF

                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else                                            // check for LF
                {
                    if (bytes == 0 && socket.isUDD())
                    {
                        cpu.data_regs[0]=Types.ERR_EOF;   
                    }
                    else
                    {
                        byte LFCheck = (byte)cpu.readMemoryByte(bytes+A1-1);
                        cpu.addr_regs[1]+=bytes;
                        cpu.data_regs[1]=D1+bytes;
                        if (LFCheck == 10)
                        {
                            cpu.data_regs[0]=0;                 // we got an LF, so we're done
                        }
                        else
                        {   
                            bytesGot+=bytes;    
                            if (bytesGot>=totalBytesToGet)      // did we as much as the buffer allowed?
                                 cpu.data_regs[0]=Types.ERR_BFFL;// yes, but no LF, so buffer is full
                            else
                                cpu.data_regs[0]=Types.ERR_NC;  // no, operation not complete yet
                        }
                    }
                }
                break;
                
            case IOB_FMUL:                                      // get multiple bytes into buffer at (a1), d2.w = buffer size
                                                                // remember this may be called several times for one operation if NC
                if (cpu.data_regs[3]==0)
                    bytesGot = 0;
                else
                    bytesGot=D1&0xffff;                         // nbr of bytes already got (if not first call)
                totalBytesToGet = D2&0xffff;                    // total nbr of bytes to get
                if (bytesGot>=totalBytesToGet)               
                {
                    cpu.data_regs[0]=0;
                    break;                                      // we got all there was to get
                }
                toGet = totalBytesToGet -bytesGot;              // I still need this many
                if (toGet> UDPHandler.MAX_PACKET_SIZE)
                    toGet = UDPHandler.MAX_PACKET_SIZE;
                bytes= socket.receive(cpu,toGet,false,null);    // get this many bytes into (A1)
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else
                {
                    if (bytes == 0 && socket.isUDD())
                    {
                        cpu.data_regs[0]=Types.ERR_EOF;   
                    }
                    else
                    {
                        bytesGot+=bytes;       
                        cpu.addr_regs[1]+=bytes;
                        cpu.data_regs[1]=D1+bytes;
                        if (bytesGot>=totalBytesToGet)          // did we get everything?
                             cpu.data_regs[0]=0;
                        else
                            cpu.data_regs[0]=Types.ERR_NC;      // no, operation not complete
                    }
                }
                break;
                   
            case IOB_SBYT:                                      // send one byte, in D1.B
                D1 = socket.send1Byte((byte)cpu.data_regs[1]);
                if (D1>0)
                    cpu.data_regs[0]=0 ;                     
                else
                    cpu.data_regs[0]=D1;
                break;
             
            case IOB_SMUL:                                      // send multiple bytes, D2.w = nbr of bytes to send
                // this is parcelled up - do not send more than MAX_PACKET_SIZE bytes a pop
                // if more is to be sent, adjust A1 and D1 accordingly and return with error NC 
                int bytesSent;                                  // d1 w = nbr of bytes alreaby send
                if (cpu.data_regs[3]==0)
                    bytesSent = 0;
                else
                    bytesSent=D1&0xffff;                        // nbr of bytes already sent (if not first call)
                int totalBytesToSend = D2&0xffff;               // total nbr of bytes to send
                if (totalBytesToSend<= bytesSent)               
                {
                    cpu.data_regs[0]=0;
                    return;                                     // we sent all there was to send
                }
                int remaining = totalBytesToSend - bytesSent;   // still need to send this many
                int toSend;
                //if (remaining<=UDPHandler.MAX_PACKET_SIZE)      // can I send what's remaining at once?
                    toSend = remaining;                         // yes
               // else
               //     toSend=UDPHandler.MAX_PACKET_SIZE;
                bytes = socket.send(cpu,toSend,null);                    
                
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else
                {
                    remaining-=bytes;       
                    cpu.addr_regs[1]+=bytes;
                    cpu.data_regs[1]=D1+bytes;
                    if (remaining<=0)                           // did we send everything?
                         cpu.data_regs[0]=0;
                    else
                        cpu.data_regs[0]=Types.ERR_NC;          // no, operation not complete
                }
                break;         
           
            case IOF_LOAD:                                      // get D2.L bytes from socket
            case IP_RECV:                                       // RECV DOES NOT ALLOW PEEK  HERE!!!!
                if (cpu.data_regs[3]==0)
                    bytesGot = 0;
                else
                    bytesGot=D1;                                // nbr of bytes already got (if not first call)
                 if (bytesGot>=D2)               
                {
                    cpu.data_regs[0]=0;
                    break;                                      // we got all there was to get
                }
                toGet = D2 - bytesGot;                          // I still need this many
                if (toGet> UDPHandler.MAX_PACKET_SIZE)
                    toGet = UDPHandler.MAX_PACKET_SIZE;
                bytes= socket.receive(cpu,toGet,false,null);    // get this many bytes into (A1)
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else
                {
                    if (bytes == 0 && socket.isUDD())
                    {
                        cpu.data_regs[0]=Types.ERR_EOF;   
                    }
                    else
                    {
                        bytesGot+=bytes;       
                        cpu.addr_regs[1]+=bytes;
                        cpu.data_regs[1]+=bytes;
                        cpu.data_regs[0]=0;
                        if (what == IOF_LOAD && bytesGot< D2)   // did we get everything if LOAD?
                            cpu.data_regs[0]=Types.ERR_NC;      // no, so operation not complete
                    }
                }
                break;
                
            case IP_RECVFM:                                     // RECV DOES NOT ALLOW PEEK  HERE!!!!
                if (D2> UDPHandler.MAX_PACKET_SIZE)
                    D2 = UDPHandler.MAX_PACKET_SIZE;
                bytes= socket.receiveFrom(cpu,D2);              // get this many bytes into (A1)
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else
                {
                 //   cpu.addr_regs[1]+=bytes;
                    cpu.data_regs[1]=16;
                    cpu.data_regs[0]=bytes;
                }
                break;
                
             case 0x80:
                D1=D1;       
            case IOF_SAVE:   
            case IP_SEND:                                       // send bytes to socket -same as IOF.SMUL, but D1 and D2 =LWs
                if (cpu.data_regs[3]==0)
                    bytesSent = 0;
                else
                    bytesSent=D1;                               // nbr of bytes already sent (if not first call)
                totalBytesToSend = D2;                          // total nbr of bytes to send
                if (totalBytesToSend<= bytesSent)               
                {
                    cpu.data_regs[0]=0;
                    return;                                     // we sent all there was to send
                }
                remaining = totalBytesToSend - bytesSent;       // still need to send this many
   //                 if (remaining<=UDPHandler.MAX_PACKET_SIZE)  // can I send what's remaining at once?
                    toSend = remaining;                         // yes
       //         else
      //              toSend=UDPHandler.MAX_PACKET_SIZE;
                bytes = socket.send(cpu,toSend,null);                    
                
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else
                {
                    remaining-=bytes;       
                    cpu.addr_regs[1]+=bytes;
                    cpu.data_regs[1]=D1+bytes;
                    if (remaining<=0)                           // did we send everything?
                         cpu.data_regs[0]=0;
                    else
                        cpu.data_regs[0]=Types.ERR_NC;          // no, operation not complete
                }
                break;
                
            case IP_SENDTO:
                int pointer=cpu.readMemoryLong(cpu.addr_regs[2]);
                int length=cpu.readMemoryLong(cpu.addr_regs[2]+4);
                 if (cpu.data_regs[3]==0)
                    bytesSent = 0;
                else
                    bytesSent=D1;                               // nbr of bytes already sent (if not first call)
                totalBytesToSend = D2;                          // total nbr of bytes to send
                if (totalBytesToSend<= bytesSent)               
                {
                    cpu.data_regs[0]=0;
                    return;                                     // we sent all there was to send
                }
                remaining = totalBytesToSend - bytesSent;       // still need to send this many
                if (remaining<=UDPHandler.MAX_PACKET_SIZE)      // can I send what's remaining at once?
                    toSend = remaining;                         // yes
                else
                    toSend=UDPHandler.MAX_PACKET_SIZE;
                bytes = socket.sendTo(cpu,pointer,toSend);
                if (bytes<0)
                {
                    cpu.data_regs[0]=bytes;                     // error code
                }
                else
                {
                    remaining-=bytes;       
                    cpu.addr_regs[1]+=bytes;
                    cpu.data_regs[1]=D1+bytes;
                    if (remaining<=0)                           // did we send everything?
                         cpu.data_regs[0]=0;
                    else
                        cpu.data_regs[0]=Types.ERR_NC;          // no, operation not complete
                }
                break;
            
            case IP_BIND :
                cpu.data_regs[0]=  socket.bind(cpu);
                break;  
                
            case IP_CONNECT :
                cpu.data_regs[0]=  socket.connect(cpu);
                break;
                
            case IP_GETHOSTNAME:
            {
                String localhostname = getHName();
                if (localhostname!=null)
                    writeCString (cpu,cpu.addr_regs[1],localhostname,cpu.data_regs[2]&0xfff);
                else
                    cpu.data_regs[0]=Types.ERR_FDNF;
            }
            break;  
                
            case IP_GETHOSTBYNAME:    
                /* ON EXIT
                 (Simplified) structure pointed to by A2

                0   pointer to A
                4   pointer to C
                8   long    type (2)
                12  long    4
                16   
                
                A       pointer to B
                B       pointer to inet address (4 bytes if ipv4)
                        B is followed by a long word of 0
                
                C       pointer to name
                */
                try
                {
                    int A2 = cpu.addr_regs[2];
                    String name =readCString(cpu,cpu.addr_regs[1]);
                    java.net.InetAddress addr=java.net.InetAddress.getByName(name);
                    
                    byte[] rawaddr=addr.getAddress();
                    cpu.writeMemoryLong(A2+8,2);                //type
                    cpu.writeMemoryLong(A2+12,4);     
                    int temp=A2+16;  
                    cpu.writeMemoryLong(A2,temp);               // pointer to A
                    cpu.writeMemoryLong(temp,temp+12);           // ptr to B into A
                    temp+=12;
                    cpu.writeMemoryLong(temp,temp+8);           // ptr to value into B
                    cpu.writeMemoryLong(temp+4,0);              // long word of 0 after B
                    bytes = temp+8;
                    temp=(rawaddr[0]&0xff)<<24 ;
                    temp+=((rawaddr[1]&0xff)<<16);
                    temp+=((rawaddr[2]&0xff)<<8);
                    temp+=(rawaddr[3]&0xff);
                    cpu.writeMemoryLong(bytes,temp);
                    temp=bytes + 4;                             // next free space
                    
                    bytes = writeCStringUnsafe (cpu,temp,name);
                    bytes+=1+temp;
                    bytes &=0xfffffffe;
                    
                    cpu.writeMemoryLong(A2+4,bytes);               // pointer to C
                    cpu.writeMemoryLong(bytes, temp);           // pointer to name
                } 
                catch (java.net.UnknownHostException ex)
                {
                    socket.setLastError (4);
                    cpu.data_regs[0]=Types.ERR_FDNF;
                }
                break;
                
            case IP_GETPEERNAME:
                cpu.data_regs[0]=socket.getRemote(cpu);
                break;
                
            case IP_H_ERRNO:
                cpu.data_regs[0]=0;
                cpu.data_regs[1]=socket.getLastError();
                break;
                
     //!!!!! this is a really unsafe operation: the address for the text to be returned is in A1 but NO MAXIMAL LENGTH IS GIVEN!!!! Stupid *!& C stuff       
            case IP_H_STRERROR:
                String s = IPHandler.IP_ERROR[socket.getLastError()];
                writeCStringUnsafe (cpu,cpu.addr_regs[1],s);
                cpu.data_regs[0]=0;
                break;
            
            case IP_ERRNO:
                cpu.data_regs[0]=0;
                cpu.data_regs[1]=this.lastError;
                break;          
                
    /*        case 0x7f:
                cpu.data_regs[0]=-10;
                socket.disconnect();
                break;
        */        
            default:
                socket.setLastError (14);
                cpu.data_regs[0]=Types.ERR_NIMP;                // what is that? whatever it is, it is not implemented for this device
                break;
        }
        this.lastError=socket.getLastError();
     }
    
    /**
     * Gets the socket.
     * 
     * @param key which socket to get.
     * 
     * @return the socket retrieved, null if not found.
     */
    public UDPSocket getSocket(int key)
    {
        return this.sockets[key];
    }
    
     /**
     * Writes a null terminated string to memory without know how much mem I have.
     *
     * Note : This is an awful hack, but: Sometimes I DON'T KNOW THE LENGTH OF THE BUFFER I'M WRITING TO,
     * as the interface doesn't give me that information.
     * So I just use the length of the string (+1 for the terminating 0) as maximum size of the buffer!!!!
     * Hello possible buffer overflow. Just pray, I guess.
     * 
     * @param cpu the CPU used
     * @param address where to write to
     * @param s       the string to write 
     * @return nbr of bytes written
     */
    public static final int writeCStringUnsafe(CPU.MC68000Cpu cpu,int address,String s)
    {
        return writeCString(cpu, address,s,s.length()+1);
    }
    
    /**
     * Writes a null terminated string to (a buffer in) memory.
     *
     * @param cpu the CPU used
     * @param address where to write to
     * @param s       the string to write 
     * @param maxlength max length of buffer
     * @return nbr of bytes written
     */
    public static final int writeCString(CPU.MC68000Cpu cpu,int address,String s,int maxlength)
    {
        smsqmulator.SmsqeString.writeSmsqeString(address,s+"\0",false,maxlength,cpu);
        int bytes = s.length()+1;
        if (bytes>(maxlength))                                  // if string written was to big for my buffer, set last char to 0
        {                                                       // so that C has the 0 terminated string it expects
            bytes=maxlength;
            cpu.writeMemoryByte(address+bytes-1, 0);
            cpu.data_regs[0]=Types.ERR_OVFL;
        }
        else
        {
            cpu.data_regs[0]=0;
        }
        return bytes;
    }
    
    /**
     * Reads a "C" type string from memory and converts it to a Java String, with SMSQE char conversion.
     * 
     * Note: potentially, this might cause the StringBuilder to overflow if no "0" is found anywhere, but I'll get an out of mem bounds error first.
     * 
     * @param cpu   the cpu with the memory array.
     * @param address where to read string in the memory.
     * 
     * @return the corresponding String.
     */
    public static final String readCString(CPU.MC68000Cpu cpu,int address)
    {
        StringBuilder p=new StringBuilder(100);
        byte b;
        while((b=(byte)cpu.readMemoryByte(address++))!=0)
        {
            p.append(smsqmulator.SmsqeString.convertToJava(b));
        }
        return p.toString();
    }
    
    /**
     * Try to get the host name.
     * 
     * @return name or null if problem.
     */
    private static String getHName ()
    {
        try
        {
            return java.net.InetAddress.getLocalHost().getHostName();// try obvious solution
        }
        catch (Exception u)
        { /*NOP*/}                                              // if problem use code below

        String host = System.getenv("COMPUTERNAME");
        if (host != null)
            return stripSite(host);
        return stripSite(System.getenv("HOSTNAME"));
    }
    
    /**
     * Strips the word ".site" from the end of a string.
     * 
     * @param s 
     * 
     * @return intiial string stripped of the word ".site" at the end, if it was there.
     */
    private static String stripSite(String s)
    {
        if (s==null)
            return null;
        return (s.toLowerCase().endsWith(".site") && s.length()>5)?s.substring(0,s.length()-5):s;
    }
   
    
    /**
     * Get the local inet v4 address as string, without it being a loopback address.
     *  
     * @return 
     */
    private static String getLocalAddress() //throws java.net.SocketException
    {
        java.util.Enumeration<java.net.NetworkInterface> ifaces;
        try
        {
            ifaces= java.net.NetworkInterface.getNetworkInterfaces();
        }
        catch (Exception e)
        {
            return "";
        }
        while( ifaces.hasMoreElements() )
        {
            java.net.NetworkInterface iface = ifaces.nextElement();
            java.util.Enumeration<java.net.InetAddress> addresses = iface.getInetAddresses();
            while( addresses.hasMoreElements() )
            {
                java.net.InetAddress addr = addresses.nextElement();
                if( addr instanceof java.net.Inet4Address && !addr.isLoopbackAddress() )
                {
                    return addr.getHostAddress();
                }
            }
        }
        return "";
    }
}

        