/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package drivers.ser;

import com.fazecast.jSerialComm.SerialPort;

/**
 * Handles all serial port IO calls from SMSQE.
 * 
 * @author and copyright Wolfgang Lenerz
 * @version  1.00 
 */
public class IOHandler
{
    private final OutputBuffer[] outBuffers;
    private final InputBuffer[] inBuffers;
    
    public IOHandler(int nbrOfSerPorts)
    {
        this.outBuffers = new OutputBuffer[nbrOfSerPorts] ;
        this.inBuffers = new InputBuffer[nbrOfSerPorts] ;
        for (int i = 0; i<nbrOfSerPorts;i++)
        {
            this.outBuffers[i]=new OutputBuffer(512);
            this.inBuffers[i]=new InputBuffer(2048);
        }
    }
    
    /**
     * Handle all the allowed serial IO operations
     * @param cpu           CPU used (sp = operation type, old D0)
     * @param serPort       serial port to handle the operations
     * @param portNumber    (SMSQE) port number
     * 
     * @return  SMSQE error code (0 = no error).
     */
    public int handleIO (CPU.MC68000Cpu cpu, SerialPort serPort, int portNumber)
    {
        int ioType = cpu.readMemoryLong(cpu.addr_regs[7]);      // IO operation wished
        cpu.addr_regs[7]+=4;                                    // pop D0 off the stack
        int D1 = cpu.data_regs[1];                              // nbr of bytes already fetched/sent
        int D2 = cpu.data_regs[2];                              // total nbr of bytes
        int D7 = cpu.data_regs[7];                              // translate & other flags   (<0 do translate)                     
        int A0 = cpu.addr_regs[0];                              // channel definition block  
        int A1 = cpu.addr_regs[1];                              // data pointer
        int A3 = cpu.addr_regs[3];                              // device definition block
        int error = 0;                                          // preset no error occurred
        
        switch (ioType)                                         // for values of ioType, see SMSQE trap#3 codes
        {
            case 0:                                             // test presence of byte
            case 1 :                                            // get one byte
                int result = this.inBuffers[portNumber].testOrGetByte(serPort, ioType==1);
                if (result<0)
                    error= result;
                else
                {   
                    cpu.data_regs[1]=(cpu.data_regs[1]&0xffffff00) | result;     // set LSByte
                }
                break;
                
            case 2 :                                            // fetch a line ending in LF
                error = this.inBuffers[portNumber].fetchLine(serPort, cpu, D2-D1);
                break;
           
            case 3 :                                            // fetch multiple bytes
                D2 &= 0x7fff;                                   // no break here, fall through!
            case 0x48:                                          // load file
                error = this.inBuffers[portNumber].fetchMultipleBytes(serPort, cpu, D2-D1);
                break;
                
            case 5 :                                            // send one byte
                error = this.outBuffers[portNumber].sendOneByte(serPort, cpu, (byte) D1,D7);//(byte) D1, D7, cpu); 
                break; 
                
            case 6 :                                            // send untranslated bytes
            case 7 :                                            // send multiple bytes (translated)                
                D2 &= 0x7fff;                                   // must be +ve word value
                cpu.data_regs[2]=D2;
            case 0x49:                                          // save file
                if (ioType == 0x49 || ioType == 6)              // remove translation flag for these IO types;
                    D7&=0x7fffffff;                             // 
                error = sendMultipleBytes(this.outBuffers[portNumber],serPort, cpu, A1, D1 , D2, D7);
                break;
            
            case 0x46:                                          // set file header
                error=sendFileHeader(this.outBuffers[portNumber],serPort, cpu, A1, D1);
                break;
                
            case 0x47:                                          // read file header
                error= readFileHeader( this.inBuffers[portNumber], serPort, cpu);
                break;
                
            default:                                            // nothing else is implemented
                error =  smsqmulator.Types.ERR_IPAR;
                break;
        }
        return error;
    }
    
    /**
     * Get a file header. Over the serial port, this is the usual 14 bytes, preceded by $FF.
     * @param buffer    buffer for the serial port
     * @param serPort   serial port whence to get read the file header
     * @param cpu
     * @return          0 or SMSQE error code.
     */
    private int readFileHeader(InputBuffer buffer, SerialPort serPort, CPU.MC68000Cpu cpu)
    {
        cpu.data_regs[2]=15;                                    // 1 announcement byte + 14 bytes header
        if ( cpu.data_regs[1]==0)                               // first time this is called?
        {                                                       // yes
            int error = buffer.testOrGetByte(serPort, false);   // fetch header announcement byte, but don't consume it yet (i.e. leave in buffer - test byte)
            if (error<0)                                        // problem when fetching the byte
                return error;                                   // kick it upstairs
            if (error!=0xff)                                    // really header anouncement byte?
                return smsqmulator.Types.ERR_IPAR;              // no, this wasn't the correct byte
            error = buffer.testOrGetByte(serPort, true);        // consume that byte now
            if (error<0)                                        // problem when fetching the byte (how can that be?????)
                return error;                                   // 
            cpu.data_regs[1] = 1;                               // show that one byte was fetched
        }
        return buffer.fetchMultipleBytes(serPort, cpu, 15 - cpu.data_regs[1]); // get the remaining bytes
    }
    
    /**
     * 
     * Send several untranslated bytes.
     * 
     * @param buffer    buffer doing the actual sending
     * @param serPort   whither to send the data
     * @param cpu       CPU with memory containing bytes to send
     * @param A1        whence in memory to take bytes (updated here)
     * @param D1        nbr of bytes already sent (on subseqent calls) (updated here)
     * @param D2        total bytes to be sent
     * 
     * @return 0 or error NC if all bytes couldn't be sent. This sets regs A1,D1 to their appropriate values
     */
    private int sendMultipleBytes(OutputBuffer buffer, SerialPort serPort, CPU.MC68000Cpu cpu, int A1, int D1, int D2, int D7)
    {
        int bytesSent = buffer.sendBytes(serPort, D2-D1, cpu, A1, D7);
        D1+=bytesSent;
        cpu.data_regs[1]=D1;                                    // update nbr of bytes actually sent
        cpu.addr_regs[1]+=bytesSent;                            // update pointer
        return (D1==D2)?0:smsqmulator.Types.ERR_NC;
    }
    
    /**
     * 
     * Send file header: 0xff followed by 14 bytes.
     * 
     * @param buffer    buffer doing the actual sending
     * @param serPort   whither to send the data
     * @param cpu       CPU with memory containing bytes to send
     * @param A1        whence in memory to take bytes (updated here)
     * @param D1        nbr of bytes already sent (on subsequent calls) (updated here)
     * @param D2        total bytes to be sent
     * 
     * @return 0 or error NC if all bytes couldn't be sent. This sets regs A1,D1 to their appropriate values
     */
    private int sendFileHeader(OutputBuffer buffer, SerialPort serPort, CPU.MC68000Cpu cpu, int A1, int D1)
    {
        int error;
        cpu.data_regs[2]=15;                                    // 1 announcement byte + 14 bytes header
      
        if (D1==0)                                              // first time around?
        {
            error = buffer.sendOneByte(serPort, cpu, (byte) -1,0);
            if (error!=0)
                return error;
            cpu.data_regs[1]=D1=1;   
        }
        return sendMultipleBytes(buffer, serPort, cpu, A1, D1, 15,0);
    }
 
    /**
     * Empty in and out buffers for a port (out buffer :  try to send everything).
     * 
     * @param serPort           port in question
     * @param portNumber        number of this port
     */
    public void emptyBuffers(SerialPort serPort, int portNumber)
    {
        this.outBuffers[portNumber].checkRoom(serPort);
        this.inBuffers[portNumber].clearBuffer();
    }
}
