/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package drivers.ser;

import com.fazecast.jSerialComm.*;

/**
 * This is a buffer that is filled in from the serial port and then input to SMSQE.
 * This is a buffer, not a queue.
 * 
 *  The head of the queue is that element which would be removed by a call to remove() or poll(). 
 *  In a FIFO queue, all new elements are inserted at the tail of the queue.
 *
 * @author and copyright Wolfgang Lenerz
 * @version  1.00 
 */
public class InputBuffer
{
    private final byte[]buffer;                                 // buffer to store data
    private int dataStart;                                      // head of the queue
    private final int size;                                     // buffer size
    private int dataEnd;                                        // up to where the buffer is filled
    private final static int SYS_XTAB = 0x146;                  // SMSQE pointer to translate table (offet in sysvars)
    private boolean EOF;                                        // true if end of file (ie no carrier detect)
    
    public InputBuffer (int psize)
    {
        this.buffer = new byte[psize];
        this.size = psize;
        this.dataStart=0;
        this.dataEnd = 0;
    }
    
    /**
     * Get data from serial port into buffer.
     * 
     * @param serPort 
     *
     */
    private void getDataIntoBuffer(SerialPort serPort)
    {
        if (! isEmpty())
            return;
        int bytesGot = serPort.readBytes(this.buffer, this.size);
        if (bytesGot>0)
        {
            this.dataEnd=bytesGot;
            this.dataStart=0;
        }
        else if (bytesGot<0)
            this.EOF=!serPort.getDCD();
    }
    
    private boolean isEmpty()
    {
        return dataEnd==0;
    }
    
    /**
     *  Checks whether the end of the buffer is reached, resets pointers if this is so.
     */
    private void checkEndReached()
    {
        if (this.dataStart==this.dataEnd)
        {
            this.dataStart=0;
            this.dataEnd=0;
        }
    }
    
    public void clearBuffer()
    {
        this.dataStart=0;
        this.dataEnd=0;
    }
    /**return
     * Tests to see whether there is a byte available.If yes, returns this byte (as +ve int) , but does not take it out of the buffer.
     * 
     * @param serPort serial port for this buffer
     * @param get
     * 
     * @return the byte as +ive int, or -SMSQE error codes (nc or eof) if there is no byte available.
     */
    public int testOrGetByte(SerialPort serPort, boolean get)
    {
        if (isEmpty())
        {
            getDataIntoBuffer(serPort);                         // my buffer is empoty, get more from ser port
            if (isEmpty())                                      // no bytes obtained from serPort
            {
                return this.EOF ? smsqmulator.Types.ERR_EOF : smsqmulator.Types.ERR_NC;
            }
        }
        int result= (this.buffer[this.dataStart] &0xff);
        if (get)
        {
            this.dataStart++;
            checkEndReached();
        }
        return result;
    } 
    
    /**
     * Fetch multiple bytes.
     * 
     * @param serPort       serial port whence to get bytes
     * @param cpu
     * @param totalToFetch  total number of bytes (still) to fetch.
     * 
     * @return SMSQE error code.
     */
    public int fetchMultipleBytes(SerialPort serPort,CPU.MC68000Cpu cpu, int totalToFetch)
    {
        if (totalToFetch == 0)
            return 0;
        getDataIntoBuffer(serPort);
        if (isEmpty())
        {
            return this.EOF ? smsqmulator.Types.ERR_EOF : smsqmulator.Types.ERR_NC;
        }
        int toGet= this.dataEnd-this.dataStart;
        if (toGet> totalToFetch )
            toGet=totalToFetch;
        int A1 = cpu.addr_regs[1];                              // bytes go here
        for (int i=0;i<toGet;i++)
        {
            cpu.writeMemoryByte(A1++, this.buffer[this.dataStart++]);
        }
        checkEndReached();
        cpu.data_regs[1]+=toGet;
        cpu.addr_regs[1] = A1;
        return toGet == totalToFetch? 0: smsqmulator.Types.ERR_NC; // either not complete, or I get everything
    }
    
    /*
    The character count of a fetch a line trap includes the linefeed character ASCII $0A, if found. The length in
D2 must be positive word (i.e. no more that $7FFF). NOTE : if no LF can be found within the length of the
buffer, the trap should return error ERR.BFFL (not OVFL as previously mentioned). Whilst the number of
bytes fetched is returned in D1.W, the upper word of D1 is NOT preserved.
* Note : For IOB.FLIN, many SMSQ/E level 3 device drivers (mostly WIN and FLP) will convert a <CR><LF>
to a simple <LF>.

    */
      /**
     * Fetch multiple bytes until a LF (0xA) is found. If it is preceded by a LF, remove LF.
     * 
     * @param serPort       serial port whence to get bytes
     * @param cpu
     * @param totalToFetch  total number of bytes (still) to fetch.
     * 
     * @return SMSQE error code.
     */
    public int fetchLine(SerialPort serPort,CPU.MC68000Cpu cpu, int totalToFetch)
    {
        if (totalToFetch == 0)
            return 0;
        getDataIntoBuffer(serPort);
        if (isEmpty())
        {
            return this.EOF ? smsqmulator.Types.ERR_EOF : smsqmulator.Types.ERR_NC;
        }
        int toGet= this.dataEnd-this.dataStart;
        if (toGet> totalToFetch )
            toGet=totalToFetch;
        int A1 = cpu.addr_regs[1];                              // bytes go here
        int i;
        boolean gotLF = false;
        for (i=0;i<toGet;i++)
        {
            int byt = this.buffer[this.dataStart++];
            cpu.writeMemoryByte(A1++, byt);
            if (byt == 10)
            {
                gotLF = true;
                break;
                        /*
                        if (cpu.readMemoryByte(A1-2) == 13)             // remove CR
                        {
                            A1-=1;                                      // point after lF
                            i-=1;                                       // one byte less was read
                            cpu.writeMemoryByte(A1-1, byt);             // write LF
                        }
                        */
            }
        }
        checkEndReached();
        cpu.data_regs[1]+=i;
        cpu.addr_regs[1] = A1;
        if (gotLF)
            return 0;
        if (cpu.data_regs[1]>=totalToFetch)
            return smsqmulator.Types.ERR_BFFL;
        return smsqmulator.Types.ERR_NC;
      //  return gotLF? 0: smsqmulator.Types.ERR_NC; // either not complete, or I got everything
    }
    
   // private void getFromBuffer
}
