/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package drivers.ser;
import com.fazecast.jSerialComm.*;
/**
 * This is a buffer that is filled in by SMSQE and then output to the serial port.
 * 
 * @author and copyright Wolfgang Lenerz
 * @version  1.00 
 */
public class OutputBuffer
{
    private int dataStart;
    private int size = 1024;
    private int dataEnd;
    private final byte[]buffer;
    private final static int SYS_XTAB = 0x146;                  // SMSQE pointer to translate table (offet in sysvars)
    
    public OutputBuffer (int psize)
    {
        this.buffer = new byte[psize];
        this.size = psize;
        this.dataStart=0;
        this.dataEnd = 0;
    }
    
    /**
     * Takes bytes out of the buffer and sends them to serial port
     * 
     * @param serPort whither to send the data
     * 
     * @return nbr of bytes written
     */
    private int outputBytes(SerialPort serPort)
    { 
        int bytesWritten = serPort.writeBytes(this.buffer, this.dataEnd-this.dataStart, this.dataStart);// there is still something in the queue, send it
        this.dataStart+=bytesWritten;
        if (this.dataStart==this.dataEnd)
        {
            this.dataStart=0;
            this.dataEnd=0;
        }
        return bytesWritten;
    }
    
    /**
     * Check that queue is empty, if not (try to) empty it.
     * 
     * @param serPort where to send data.
     * 
     * @return true if queue is empty and can accept new data, false if not
     */
    public boolean checkRoom(SerialPort serPort)
    {
        if (this.dataStart==0)
            return true;                                        // I have room for more
        outputBytes(serPort);                                   // send what is still buffered
        return this.dataStart==0;                               // if this is 0 buffer is now free
    }
    
    /**
     * Send one byte to the serial port.
     * 
     * @param serPort  where to send data
     * @param cpu
     * @param b         byte to be sent
     * @param D7        flag : if bit 32 set: do translate
     * 
     * @return 0 or SMSQE error code
     */
    public int sendOneByte(SerialPort serPort, CPU.MC68000Cpu cpu, byte b, int D7)
    {
        if (!checkRoom(serPort))
            return smsqmulator.Types.ERR_NC;                    // buffer not empty
        if (D7<0 && b!=0)
        {
            byte[]txl = translateByte(cpu,b);
            this.dataEnd=txl.length;
            System.arraycopy(txl, 0, this.buffer, 0, this.dataEnd);
        }
        else
        {
            this.buffer[0]=b;
            this.dataEnd=1;
        }
        this.dataStart=0;                                       // don't use checkroom directly here (dataStart = 0 !)
        int bytesWritten = outputBytes(serPort);                // 
        return bytesWritten==1?0: smsqmulator.Types.ERR_NC;     // Ok if 1 byte sent, else tell SMSQE to try again later.
    }
    
    /**
     * Send multiple bytes to the serial port (bytes lay at (A1))
     * 
     * @param serPort               where to send data
     * @param nbrOfBytesToSend      how many bytes are to be sent
     * @param cpu                   CPU with memory containing bytes to send
     * @param A1                    whence in memory to take bytes
     * @param D7                    flag : if bit 32 set: do translate
     * 
     * @return number of bytes sent - if this is 0, nothing was sent
     */
    public int sendBytes(SerialPort serPort, int nbrOfBytesToSend, CPU.MC68000Cpu cpu, int A1, int D7)
    {
        if (!checkRoom(serPort))
            return 0;                                           // queue is still not empty
        if (D7<0)
            return sendBytesWithTranslate(serPort, nbrOfBytesToSend, cpu, A1);
        int toSend = (nbrOfBytesToSend>this.size)?this.size:nbrOfBytesToSend;
        for (int i = 0; i<toSend; i++)
        { 
            this.buffer[i]=(byte)cpu.readMemoryByte(A1++);
        }
        this.dataEnd=toSend;
        int bytesSent=outputBytes (serPort);                    // pretend that buffer is now empty (SMSQE will resend the bytes that weren't actually sent).
        return bytesSent;                                       // this many bytes were sent
    }
    
     /**
     * Send multiple bytes to the serial port, with translate.Bytes lay at (a1)
     * 
     * @param serPort               where to send data
     * @param nbrOfBytesToSend      how many bytes are to be sent
     * @param cpu                   CPU with memory containing bytes to send
     * @param A1                    whence in memory to take bytesd
     * 
     * @return number of bytes sent - if this is 0, nothing was sent
     */
    private int sendBytesWithTranslate(SerialPort serPort, int nbrOfBytesToSend, CPU.MC68000Cpu cpu, int A1)
    {
        int maxSend =40;                                        // limit nbr of bytes to send in one go, to make sure very translated byte is sent entirely (i.e. all translated 3 bytes)
        int bytesSent = 0;
        int toSend = (nbrOfBytesToSend>maxSend)?this.size:nbrOfBytesToSend;
        for (int i = 0; i<toSend; i++)
        { 
            byte[] translated = translateByte(cpu,(byte)cpu.readMemoryByte(A1++));
            System.arraycopy(translated, 0, this.buffer,this.dataStart,translated.length);
            this.dataStart+=translated.length;
            bytesSent++;
        }
        outputBytes (serPort);
        return bytesSent;                                       // THERE IS THE POSSIBILITY HERE THAT ALL BYTES WILL NOT BE SENT IF THIS WAS THE LAST BATCH (THEN smsqe WILL
                                                                // NOT TRY TO RESEND BYTES) -- THESE SHOULD THEN BE WHEN CHANNEL IS CLOSED.
    }
    
    /**
     * Does byte translation from SMSQE's translation table.
     * 
     * @param b byte to translate
     * 
     * @return an array with the translated byte(s): One byte may result in up to three translated bytes.
     */
    private byte[] translateByte(CPU.MC68000Cpu cpu, byte b)
    {
        if (b==0)
            return new byte[1];                                 // null byte (isn't translated)
        int value = (b&0xff); 
        int tTable = cpu.readMemoryLong(cpu.addr_regs[6]+OutputBuffer.SYS_XTAB); // pointer to translation table...
        int tableCopy = tTable;
        tTable += cpu.readMemoryWord(tTable+2);                 // ... now
        int res = cpu.readMemoryByte (tTable+value);            // translated byte
        if (res!=0)                                             // single replace?
            return new byte[]{(byte) res};                      // yes, done
        
        // 1 bytes translates into 3 others : get the table and read it until I find my value;
        tTable = tableCopy + cpu.readMemoryWord(tableCopy+4);   // pointer to translation table for multiple replacements
        int nbrEntries = cpu.readMemoryByte(tTable);            // nbr of entries in translattion table;
        tTable += 1;                                            // point 1st entry;
        for (int i = 0;i<nbrEntries;i++)
        {
            if (cpu.readMemoryByte(tTable) == value)
            {
                tTable++;
                byte[] result=new byte[3];
                for (int j = 0;j<3;j++,tTable++)
                    result[j]=(byte)cpu.readMemoryByte(tTable);
                return result;
            }
            tTable+=4;                                          // go to next entry in table
        }                                         
                                                                //if we get here, byte not found in translation table, so return " " (space)
        return new byte[]{32}; 
    }
    
    /*
    in iod_iob_iox_asm
        page
; put byte with translate for JS and MG ROMs
ibio_tra
ix.reg  reg     d1/d2/a1
        movem.l ix.reg,-(sp)
        cmp.b   #k.ff,d1                 ; form feed?
        bhi.s   ix_trck
        beq.s   ix_ff
        cmp.b   #k.nl,d1                 ; newline?
        bne.s   ix_trck

        cmp.w   #$200,d7                 ; <CR>?
        blt.s   ix_trck                  ; ... no

        moveq   #k.cr,d1
        cmp.w   #$400,d7                 ; <CR><LF>?
        blt.s   ix_sbyte                 ; ... no, <CR>

ix_crlf
        lea     ixb_crlf,a1
        moveq   #2,d2
        bra.s   ix_smul

ix_ff
        cmp.w   #$400,d7                 ; <CR><FF>
        blt.s   ix_trck                  ; ... no
        lea     ixb_crff,a1
        moveq   #2,d2
        bra.s   ix_smul

ix_trck
        tst.l   d7                       ; translate?
        bpl.s   ix_sbyte                 ; ... no
ix_do

    
       ix_do
        move.l  sys_xtab(a6),a1          ; magic system variable
        add.w   2(a1),a1                 ; pointer to Xlate table
        moveq   #0,d2                    ; get word in d3
        move.b  d1,d2
        beq.s   ix_sbyte                 ; ... null
        move.b  (a1,d2.w),d1             ; get xlated byte
        bne.s   ix_sbyte                 ; ... done

        move.l  sys_xtab(a6),a1          ; magic system variable
        add.w   4(a1),a1                 ; pointer to extension table
        move.b  (a1)+,d1                 ; number of entries
        bra.s   ix_elook
ix_look
        cmp.b   (a1)+,d2                 ; our character?
        beq.s   ix_sthree                ; ... yes, send next three bytes
        addq.l  #3,a1                    ; ... next character
        subq.b  #1,d1
ix_elook
        bne.s   ix_look

        moveq   #' ',d1
        bra.s   ix_sbyte                 ; not found, send space

    */
}
