/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package drivers.ser;

import com.fazecast.jSerialComm.*;
import static com.fazecast.jSerialComm.SerialPort.*;
import java.util.Arrays;

/**
 * The serial port driver interface. It allows for 4 different serial ports. 
 * The association with real comm ports must be set up before this object is initialized.
 * 
 * As I use an outside library for this, I need to conform to that. Notably, I nerd to set the port parameters here, not from SMSQE.
 * That being so, I might just as well do everything from here and not from smsqe.
 *  
 * SMSQE ser ports come in 3 flavours : input only, output only, both. See the openingIsAllowed method for peculiarities.
 * 
 * @author and copyright Wolfgang Lenerz
 * @version  1.00
 */
public class SerDriver
{
    // SMSQE interface ; Keys for comms with smsqe. If you change any of these, also change them in smsqe
    private static final int CHECK_PORT_EXISTS = 0;
    private static final int SET_BAUD = CHECK_PORT_EXISTS+1;
    private static final int OPEN_PORT = SET_BAUD+1;            // will also set parity/handshake (taken from port linkage area)
    private static final int CLOSE_PORT = OPEN_PORT+1;
    private static final int PORT_IO = CLOSE_PORT+1;
    private static final int SET_PORTNAME = PORT_IO+1;          // set host name for serial port, obtained from SMSQE
    private static final int GET_PORTNAME = SET_PORTNAME+1;     // return host name for serial port to SMSQE
    private static final int GET_ALLPORTNAMES = GET_PORTNAME+1; // return host name for serial port to SMSQE
   
    // constants defined in SMSQE (keys_serpar_prt)
    private static final int SPD_LEN  = 0xA0;                   // length of one ser linkage block in SMSQE
    private static final int SPD_PORT = 0x54;                   // offset in linkage block :  ser port nbr
    private static final int SPD_HAND = 0x8D;                   // offset in linkage area flow control: -ve hardware +ve XON/XOFF, 0=none
    private static final int SPC_PRTY = 0x1A;                   // offset in channel block (word) parity 0 none, 2 odd, 4 even, 6 mark, 8 space
    private static final int SPC_INOUT= 0x22;                   // offset in channel block (word) direction 3=both, 1=in, 2=out 
    
    private static final int UNUSED_PORT_DIRECTION = -4;        // direction of unused ports (the last 2 bits are not set)
    private static final int INPUT_SET  = 1;                    // bit 0 set if channel allows input
    private static final int OUTPUT_SET = 2;                    // bit 1 set if channel allows output

    private static final int MAX_PORTS = 4;                     // how many serial ports are allowed
//    private static final int MAX_BAUDRATE = 115200;             // artificially set maximum baud rate
    private final SerialPort[] serPorts=new SerialPort[SerDriver.MAX_PORTS];// the serial port objects themselves
    private final String[] portNames =  new  String[SerDriver.MAX_PORTS]; // names for these ports
    private final int[] bauds = new int[SerDriver.MAX_PORTS];   // baud rates for ports (SMSQE might set baud rates before port is open)
    private final IOHandler ioHandler = new IOHandler(SerDriver.MAX_PORTS);// handles IO calls
    private final int []directions = new int[MAX_PORTS];        // what directions does the port allow
    private final int [] nbrOfPorts=new int[MAX_PORTS];         // how many SMSQE ports are open to one SerialPort
    
    /**
     * Create the object with an existing set of names (MUST be a 4 element String array).
     * 
     * @param portnames String array with Driver.MAX_PORTS (=4) elements (anything over that is ignored) - if this is null, empty names ("") will be created.
     */
    public SerDriver (String[] portnames)
    {
        setNames(portnames);
        setDefaultBaudRate();
        Arrays.fill(this.directions, -4);
    } 
   
    /**
     * Create the object with no preset set of names - get the names from the machine.
     * 
     * I use the names of the serial ports present in the machine. 
     */
    public SerDriver ()
    {
        SerialPort[] ports = SerialPort.getCommPorts();
        int leng = ports.length;
        if (leng>SerDriver.MAX_PORTS)
            leng = SerDriver.MAX_PORTS;
        int i;
        String intern = System.getProperty("os.name").toLowerCase(); // on linux (and mac) I need to add /dev/ before the names
        boolean addDevicePath =  (intern.contains("linux")||intern.contains("mac"));
        for ( i = 0;i<leng;i++) 
        {
            SerialPort port = ports[i];
            this.portNames[i]=addDevicePath? "/dev/"+port.getSystemPortName() : port.getSystemPortName();
        }   
        if (i<SerDriver.MAX_PORTS)
        {
            for (;i<SerDriver.MAX_PORTS;i++)
            {
                 this.portNames[i]="";
            }
        }
        setDefaultBaudRate();
        Arrays.fill(this.directions, -4);
    }
    
    /**
     * Create the object with data to be taken from an ini file.
     * 
     * @param inifile whence the data comes from.
     */
    public SerDriver (inifile.IniFile inifile)
    {
        this();
        String mainPart = "Ser";
        String [] names = new String [SerDriver.MAX_PORTS];
        for (int i=0;i<SerDriver.MAX_PORTS;i++)
        {
            names[i]=inifile.getTrimmedOptionValue(mainPart+i);
        }
        setNames(names); 
        setDefaultBaudRate();
    }  
    
    
    /**
     * Set names of ports from the supplied string array.
     * 
     * @param portnames String array with MAX_PORTS (=4) elements (anything over that is ignored) - if this is null, empty names ("") will be created.
     */
    public final void setNames(String[] portnames )
    {
        int i=0;
        int leng;
        if (portnames!=null)                                    // there are some port names
        {
            leng= portnames.length;
            if (leng>SerDriver.MAX_PORTS)
                leng = SerDriver.MAX_PORTS;
            for (i = 0;i<leng;i++) 
            {
                this.portNames[i]=portnames[i];
            }    
        }
        if (i<SerDriver.MAX_PORTS)
        {
            for (;i<SerDriver.MAX_PORTS;i++)
            {
                 this.portNames[i]="";
            }
        }
    }
    
    
    /**
     * Try to open the port and return success status.
     * 
     * @param portNumber which port (0...3) to open.
     * @param cpu
     * 
     * @return 0 if port opened OK, else smsqe error code.
     */
    public int openPort(int portNumber, CPU.MC68000Cpu cpu)      
    {
        if (portNumber<0 || portNumber>=SerDriver.MAX_PORTS)
            return smsqmulator.Types.ERR_IPAR;                 
        if (this.portNames[portNumber]==null || this.portNames[portNumber].isEmpty())
            return smsqmulator.Types.ERR_IPAR;
        
        int newDirections = makeDirection (cpu.data_regs[5]);   // smsqe direction is in D5 : -1 out only, 0 in and out, 1 in only
        int currentDirections = this.directions[portNumber];    // the currently allowed directions
        SerialPort serPort = this.serPorts[portNumber];         // may be null
        boolean resetParameters = (newDirections == currentDirections); // will be set to true if I need to change parameters
        if (serPort!=null)                                      // port already open, can I use new one?
        {
            if (!(openingIsAllowed(newDirections,currentDirections)))
                return smsqmulator.Types.ERR_FDIU;              // can't open additional or complementary port.
            switch(newDirections)
            {
                case OUTPUT_SET:                                // currently out only
                    currentDirections|=INPUT_SET;               // allow also input
                    cpu.writeMemoryByte(cpu.addr_regs[0]+SPC_INOUT, newDirections); // keep  track of what kind of channel this is.
                    break;
                    
                case INPUT_SET:                                
                    currentDirections|=OUTPUT_SET;
                    resetParameters = false;
                    break;
            }
            cpu.writeMemoryByte(cpu.addr_regs[0]+SPC_INOUT, newDirections);// keep  track of what kind of channel this is.
            resetParameters = false;                    // don't change port parameters
        }
        else
        {
            try
            {
               this.serPorts[portNumber] = SerialPort.getCommPort(this.portNames[portNumber]);
               serPort = this.serPorts[portNumber];
            }
            catch (SerialPortInvalidPortException e)
            {
                return smsqmulator.Types.ERR_NIMP;
            }
            if (!serPort.openPort())
                return smsqmulator.Types.ERR_NIMP;
            resetParameters = true;
            switch (newDirections)
            {
                case -1:
                    currentDirections = OUTPUT_SET;       
                    break;
                case 0:
                    currentDirections = OUTPUT_SET + INPUT_SET; 
                    break;  
                case 1:
                    currentDirections = INPUT_SET;                   
                    break;
            }
        }
        this.directions[portNumber] = currentDirections;
        this.nbrOfPorts[portNumber]++;
        if ((currentDirections & INPUT_SET)!=0)
        {
            serPort.removeDataListener();
            serPort.addDataListener(new SerialPortDataListener() 
            {
                @Override
                public void serialEvent(SerialPortEvent event) 
                {
                 // NOP!
                }

                @Override
                public int getListeningEvents() 
                { 
                    return SerialPort.LISTENING_EVENT_DATA_AVAILABLE;  
                }
            });
        }
        if (resetParameters)
        {
            setParameters(portNumber,cpu);

          //  serPort.setComPortParameters(MAX_BAUDRATE,8,1, 0, false);                 // default parameters
        }
        return 0;
    }
   
    /**
     * Check whether I can open a complementary or additional port. 
     * Opening an additional (i.e. SAME KIND OF) port for an existing port is only allowed if this is an STX port.
     * If it is not the same kind of port, then the open port must be uni-directional and only the complement will be allowed.
     * (i.e. if STX: SRX may be opened, if SRX: STX may be opened.)
     * Anything else will be rejected.
     * 
     * @param newDirections     new directions of port wished.
     * @param currentDirections current direction of open port
     * .
     * @return true if port may be opened.
     */
    private boolean openingIsAllowed(int newDirections, int currentDirections)
    {
        if (newDirections == currentDirections)             
            return (currentDirections== OUTPUT_SET);        // I can re-open an STX type of port
         
        switch (newDirections)
        {
            case -1:                                        // currently STX output only
                return (currentDirections == 1);            // only allow complementary   
            
            default:                                        // there should be no other value than the cases, though!!!
            case 0:                                         // currently SER input + output
                return false;                               // can't open a different port at all            

            case 1:                                         // currently SRX input only
                return (currentDirections == -1);           // only allow complementary 
        }                                      
    }
    
    
    /**
     * Close the port with that port number.
     * 
     * @param portNumber number of port to close.
     */
    private void closePort(int portNumber)
    {
        this.nbrOfPorts[portNumber]--;   
        if (this.serPorts[portNumber]!=null)    
        { 
            this.serPorts[portNumber].removeDataListener(); 
            this.serPorts[portNumber].closePort();
            this.ioHandler.emptyBuffers(this.serPorts[portNumber], portNumber);
        }
        this.serPorts[portNumber]=null;
        this.directions[portNumber] = -4;
    }
        
    /**
     * Sets the default baud rate for all ports to 9600
     */
    private void setDefaultBaudRate()
    {
        for (int i = 0;i<SerDriver.MAX_PORTS;i++)
            this.bauds[i]=9600;
    }
    
    /**
     * Handle all calls from SMSQE.
     * This sets D0 as per SMSQE code requirements.
     * 
     * @param cpu the CPU used.
     */
    public void handleSMSQECalls(CPU.MC68000Cpu cpu)
    {
        int error=0;                                            // preset no error
        switch (cpu.data_regs[0])                               // D0 contains the action ocde
        {
            case CHECK_PORT_EXISTS:                             // check port exists or highest port nbr
                int toCheck = cpu.addr_regs[4];                 // (D0 is in A4) what to check : if 0, get highest nbr, else check port n
                if (toCheck<0 || toCheck >=MAX_PORTS+1)         // SMSQE ser ports numbering starts at 1
                    error=smsqmulator.Types.ERR_IPAR;           // there are no negative ports = THIS SHOULD NEVER HAPPEN
                else if (toCheck == 0)                          
                {
                    error=SerDriver.MAX_PORTS;                  // return highest port nbr (as ERROR!!!! as per SMSQE requirement)
                }
                else                                            // check whether a particular port exists
                {
                    toCheck--;
                    if (this.portNames[toCheck]!=null && !this.portNames[toCheck].isEmpty())
                         error=SerDriver.MAX_PORTS;             // if port exists, return any number exept 0 (! i.e. return an error)
                                                                // if port doesn't exist, return D0 as 0!!!!!
                }
                break;
                
            case SET_BAUD:                                     
                int portNumber = getPortnumber(cpu);            // returns port number or smsqe error code
                if (portNumber<0)
                {
                    error = portNumber;                         // bad port number
                    break;  
                }
                
                int baud = BaudRate.getBaudRate(cpu.data_regs[1]); // returns baud rate or smsqe error code
                if (baud<0)  
                    error = baud;                               // port can't handle this baud rate
                else
                {
                    this.bauds[portNumber]=baud;
                    if (this.serPorts[portNumber]!=null)   
                    {
                        if (!this.serPorts[portNumber].setBaudRate(this.bauds[portNumber]))
                        {
                            error  = smsqmulator.Types.ERR_IPAR;  // port can't handle this baud rate
                        }
                    }
                }
                break;
                
            case OPEN_PORT:                                
                portNumber = getPortnumber(cpu);
                if (portNumber<0)
                    error = portNumber;                         // bad port number
                else
                    error = openPort(portNumber,cpu);   
                break;
               
            case CLOSE_PORT:                                    // a4 = old d0!
                portNumber = getPortnumber(cpu);
                if (portNumber>-1)
                    closePort(portNumber);
                error = cpu.addr_regs[4];                       // check status on old d0  !!!!
                break;
                
            case PORT_IO:
                portNumber = getPortnumber(cpu);
                if (portNumber<0 || this.serPorts[portNumber]==null)
                    error = portNumber;
                else
                    error = this.ioHandler.handleIO(cpu, serPorts[portNumber],portNumber);
                break;
                 
            case SET_PORTNAME:                                  // on entry A1 points to param
                int A1 = cpu.addr_regs[1];
                portNumber = cpu.readMemoryLong(A1)-1;
                A1= cpu.readMemoryLong(A1+8);
                if (portNumber <0 || portNumber>=SerDriver.MAX_PORTS || cpu.readMemoryWord(A1) <= 0)
                    error = smsqmulator.Types.ERR_IPAR;
                else
                {
                    String s = smsqmulator.SmsqeString.readSmsqeString(A1,cpu);
                    closePort(portNumber);
                    this.portNames[portNumber]=s;
                }
                break;
            
            case GET_PORTNAME:
                A1 = cpu.addr_regs[1];
                portNumber =  cpu.readMemoryLong(A1)-1;
                if (portNumber <0 || portNumber>=SerDriver.MAX_PORTS)
                    error = smsqmulator.Types.ERR_IPAR;
                else
                {
                    int bufferSize=cpu.readMemoryWord(A1+6);
                    int bufferAddress = cpu.readMemoryLong(A1+8);
                    String s = this.portNames[portNumber];
                    if (s==null)
                        cpu.writeMemoryWord(bufferAddress,0);                  // there is nothing there
                    else
                    {
                        smsqmulator.SmsqeString.writeSmsqeString(bufferAddress, s, true, bufferSize,cpu);
                        if (bufferSize<s.length())
                            error = smsqmulator.Types.ERR_BFFL;
                    }
                }
                break;
                
                case GET_ALLPORTNAMES:
                    A1 = cpu.addr_regs[1];
                    int bufferSize=cpu.readMemoryWord(A1+2);
                    int bufferAddress = cpu.readMemoryLong(A1+4);
                    StringBuilder sb = new StringBuilder(bufferSize);
                    byte lf = 10;
                    for (int i=0;i<this.portNames.length; i++)
                    {   
                        String s = this.portNames[i];
                        if (s!=null)
                            sb.append(s);
                        sb.append("\n");
                    }
                    smsqmulator.SmsqeString.writeSmsqeString(bufferAddress, sb.toString(), true, bufferSize,cpu);
                    if (bufferSize<sb.length())
                        error = smsqmulator.Types.ERR_BFFL;
                    break;
                
            default:
                error = smsqmulator.Types.ERR_NIMP;
                break;
        }
        checkError(cpu,error);
    }
    
    
    /**
     * Set the parameters for a port. 
     * Once this is done, only the baud rate may be changed, else the port has to be closed and reopened in SMSQE.
     * 
     * @param portNumber    what port is concerned.
     * @param cpu           current CPU used.
     */
    private void setParameters (int portNumber,CPU.MC68000Cpu cpu)
    {
        int baud = this.bauds[portNumber];
        int parity = convertParity(cpu);
        int dataBits = (parity==0)?8:7;                         // I'm not sure about that!
        int stopBits = 1;
        this.serPorts[portNumber].setComPortParameters(baud,dataBits,stopBits, parity, false);
        
        int flowControl = convertFlowControl(cpu);
        this.serPorts[portNumber].setFlowControl(flowControl);              
    }
    
    /**
     * Converts from SMSQE parity param to jserialcomm param
     * Parity for SMSQE (at SPC_PRTY(A0))
     *   0 none
     *   2 odd
     *   4 even
     *   6 mark
     *   8 space
     * 
     * Parity for jserialcomm:
     *   0 none
     *   1 odd
     *   2 even
     *   3 mark
     *   4 space
     * 
     * @param cpu
     * 
     * @return the jserialcomm parity code.
     */
    private int convertParity (CPU.MC68000Cpu cpu)
    {
        int parity = cpu.readMemoryWord(cpu.addr_regs[0]+SPC_PRTY);
        return parity/2;
    }
    
    /**
     * Converts from SMSQE flow control parameters to jserialcomm parameters.
     * Flow control for SMSQE (byte at SPD_HAND(A3))
     *   -ve    hardware 
     *   0      none
     *   +ve    XON/XOFF
     * 
     *  Flow control for jserialcomm:
     *   none       0
     *   hardware   (FLOW_CONTROL_RTS_ENABLED | FLOW_CONTROL_CTS_ENABLED)
     *   XON/XOFF   (FLOW_CONTROL_XONXOFF_IN_ENABLED | FLOW_CONTROL_XONXOFF_OUT_ENABLED)
     * 
     * @param cpu
     * 
     * @return the jserialcomm parity code.
     */
    private int convertFlowControl(CPU.MC68000Cpu cpu)
    {
        int ctrl = cpu.readMemoryByteSigned(cpu.addr_regs[3]+SPD_HAND);
        if (ctrl<0)
            return FLOW_CONTROL_RTS_ENABLED | FLOW_CONTROL_CTS_ENABLED;
        if (ctrl>0)
            return FLOW_CONTROL_XONXOFF_IN_ENABLED | FLOW_CONTROL_XONXOFF_OUT_ENABLED;
        return ctrl;
    }
    
    /**
     * Set default parameters for a port
     * These are : 9600 bauds, 8 data bits, 1 stop bit, no parity (no flow control).
     * @param serPort 
     */
    public void setDefaultParameters(SerialPort serPort)
    {
        serPort.setComPortParameters(9600,8,1, 0, false);
        serPort.setFlowControl ( FLOW_CONTROL_DISABLED);
    }
    
    /**
     * Show SMSQE that an error occurred: set D0 and status regs.
     * @param cpu
     * @param errorCode 
     */
    public static void error(CPU.MC68000Cpu cpu, int errorCode)
    {
        cpu.data_regs[0]=errorCode;
        cpu.reg_sr &=0xff10;                                    // keep X flag, 0 the rest
        if (errorCode<0)  
            cpu.reg_sr&=~8;                                     // set N flag if negative  
    }
    
    /**
     * Show SMSQE that no error occurred: simulate "CLR.L d0" : set Z flag, set D0 to 0.
     * @param cpu 
     */
    public static void noError(CPU.MC68000Cpu cpu)
    {    
        cpu.data_regs[0]=0; 
        cpu.reg_sr &=0xff10;                                        // keep X flag, 0 the rest
        cpu.reg_sr +=4;                                         // set Z flag
    }
     /**
     * Show SMSQE that no error occurred or not, depending ton the parameter.
     * 
     * @param cpu 
     * @param toCheck return as error if this is !=0, else as no error.
     */
    public static void checkError (CPU.MC68000Cpu cpu, int toCheck)
    {
        if (toCheck==0)
            noError (cpu);
        else
            error (cpu,toCheck);
    }
    /**
     * Gets the port number.
     * 
     * @param cpu whence to get the port number (at offset SPD_PORT from (a3)).
     * 
     * @return the port number or smsqmulator.Types.ERR_NIMP if problem.
     */
    public static int getPortnumber (CPU.MC68000Cpu cpu)
    {  
        int portNumber = cpu.readMemoryWord(cpu.addr_regs[3]+SPD_PORT) -1; // port number is here
        if (portNumber<0 || portNumber>=SerDriver.MAX_PORTS)
            return smsqmulator.Types.ERR_NIMP;
        return portNumber;
    }
    
    /**
     * 
     * @return 
     */
    public static java.util.List<SerPortDefinition> getSerialPorts()
    {       
	SerialPort[] ports = SerialPort.getCommPorts();
        java.util.List<SerPortDefinition> list =new java.util.ArrayList<>(ports.length);
        for( int i = 0;i<ports.length;i++) 
        {
            SerialPort port = ports[i];
        //    list.add(new SerPortDefinition(port, "/dev/"+port.getSystemPortName(),port.getDescriptivePortName(), port.getPortDescription()));
            list.add(new SerPortDefinition(port, "/dev/"+port.getSystemPortName(),port.getDescriptivePortName(), ""));
        }
        return list;
    }
  
    /**
     * Close all ports of this driver.
     */
    public void closeAllPorts()
    {
        for (int i=0;i<SerDriver.MAX_PORTS;i++)
        {
            if (this.serPorts[i]!=null)
            {
                closePort(i);
            }
        }
    }
    
    /**
     * Set the machine name for all ports.
     * 
     * @param portNames (machine) names of the port. Should be a 4 element array, all elements beyond the fourth are ignored.
     */
    public void setAllPortnames2 (String[] portNames)
    {
        if (portNames==null)
            return;
        int size = portNames.length;
        if (size>3)
            size = 3;
        System.arraycopy(portNames, 0, this.portNames, 0, size);
    }
    
    public static int getMaxSerialPorts()
    {
        return SerDriver.MAX_PORTS;
    }
    
    /**
     * Changes the Smsqe direction code to mine                 // direction in D5 : 
     * 
     * @param smsqeDirection                                    // can be: -1= out only, 0= in and out, 1= in only
     * 
     * @return my direction
     */
    private int makeDirection (int smsqeDirection)
    {
        switch (smsqeDirection)
        {
            case -1:
                return OUTPUT_SET;
            case 0:
            default:
                return INPUT_SET+OUTPUT_SET;
            case 1:
                return INPUT_SET;
        }
    }
}
