/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package drivers.win;

/**
 *
 * @author wolf
 */
public class FAT 
{
    private final java.util.ArrayList <Integer> clusterChain;
    private final int clusterSize;
    private final java.nio.ByteBuffer driveFAT;
    
    /**
     * Creates the object and reads the drive's header and FAT.
     * The FAT starts at offset 0x40 of the very first sector. So the fist cluster (cluster 0) already is part of the FAT.
     *
     * @param ioChannel channel to the underlying file on the native file system.
     * @param warnIfQXLDriveNotCompliant 
     * @param name   drive name
     * @param number drive number (1 ... 8)
     * 
     * @throws java.io.IOException any IO exception reading the map / main dir.
     * @throws ArrayIndexOutOfBoundsException a wrong map cluster number was given.
     * @throws smsqmulator.DoNothingException badly constructed qxl.win drive
     * @throws IllegalStateException if the raFile isn't null when this object is created (how could that be?????).
     * @throws drivers.IncorrectFiletypeException
     */
        public FAT (java.nio.channels.FileChannel ioChannel, boolean warnIfQXLDriveNotCompliant,String name, int number)
                    throws  java.io.IOException, ArrayIndexOutOfBoundsException,IllegalStateException,
                            smsqmulator.DoNothingException, drivers.IncorrectFiletypeException
    {
        java.nio.ByteBuffer firstSector=java.nio.ByteBuffer.allocate(512);
        int bytesRead=ioChannel.read(firstSector);     // read the first sector (512 bytes - NOT the first cluster) of the drive
        if (bytesRead!=512 || firstSector.getInt(0)!=WinDriver.QLWA)// I must have read exactly this amount and marker must be valid
        {
           throw new drivers.IncorrectFiletypeException(ui.Localization.Texts[111]); // if this isn't possible, this can't be a true QLWA file
        }
        this.clusterSize=512*(firstSector.getShort(WinDrive.QWA_SCTG)&0xffff); // size of one cluster in bytes
        this.clusterChain= new java.util.ArrayList <Integer>(80);//cluster chain for FAT
        this.clusterChain.add(0);                        // first cluster of FAT (map) is always cluster 0
        // now find out the cluster numbers for the FAT.
        // for convenience, we'll hold the entire FAT + header in memory - we first need to find out how many clusters the FAT occupies
        // the FATs clusterchain lies at the every beginnng of the FAT.
        int cluster=firstSector.getShort(WinDrive.QWA_GMAP)&0xffff;// next cluster for map
        try                                                     // we assume that the entire FAT clusterchain fits into the first sector...
        {
            while (cluster!=0)
            {       
                this.clusterChain.add(cluster);          // one more cluster
                cluster=firstSector.getShort(WinDrive.QWA_GMAP+cluster*2)&0xffff;//get pointer to next cluster
            }
        }
        catch (Exception e)  
        {
           throw new ArrayIndexOutOfBoundsException();     // Huh????
        }
        
        // patch for badly constructed drives, like those built by older versions of qxltools. These drives do NOT have
        // cluster entries for the FAT(!!!!). They just indicate the number of sectors taken by the FAT.
        // I then presume that the FAT's sectors are all contiguous and start at sector 0.
        int nbrOfSectorsInMap=firstSector.getShort(WinDrive.QWA_SCTM)&0xffff;// that many sectors are supposed to be in the fat
        int temp=nbrOfSectorsInMap/(firstSector.getShort(WinDrive.QWA_SCTG)&0xffff);
        if (nbrOfSectorsInMap % (firstSector.getShort(WinDrive.QWA_SCTG)&0xffff)!=0)
            temp++;                                 // these are the number of clusters needed to hold the map
        if (temp!=this.clusterChain.size())         // if they both agree, all is ok, I have a valid drive map
        {   
            if (warnIfQXLDriveNotCompliant)         // but here I don't, possibly ask user what to do
            {
                java.awt.Toolkit.getDefaultToolkit().beep();
                int reply=javax.swing.JOptionPane.showConfirmDialog
                (
                    null,
                    ui.Localization.Texts[74]+name+ui.Localization.Texts[75]+"\n"+
                    ui.Localization.Texts[76]+"\n"+
                    ui.Localization.Texts[77]+"WIN"+number+"_ "+ui.Localization.Texts[78]+"\n",
                    ui.Localization.Texts[73],
                    javax.swing.JOptionPane.WARNING_MESSAGE,
                    javax.swing.JOptionPane.YES_NO_OPTION
                );
                if (reply!=javax.swing.JOptionPane.YES_OPTION)  
                {
                    throw new smsqmulator.DoNothingException();             // this signals that we should abandon trying to fix the drive
                }
            }
            for (int i = 1;i<temp;i++)
                this.clusterChain.add(i);                    // make a clusterchain representing xxx first clusters
        }
        // if we get here, we have the number of clusters occupied by the FAT
        this.driveFAT=java.nio.ByteBuffer.allocate(this.clusterSize*this.clusterChain.size());// get space for all clusters
        this.driveFAT.position(0);
       // this.driveFAT.limit(this.driveFAT.capacity());
        for (int i: this.clusterChain)                        // fill the entire clusterchain for the fat
        {
            this.driveFAT.limit(this.driveFAT.position()+this.clusterSize);// read in one cluster at a time
            ioChannel.position(i*this.clusterSize);        // position in file
            ioChannel.read(this.driveFAT);                 // read (part of) the file 
        }   
    }
        
    public short getByte(int index)
    {
        return this.driveFAT.get(index);
    }
    
    public short getShort(int index)
    {
        return this.driveFAT.getShort(index);
    }
    
    public int getInt(int index)
    {
        return this.driveFAT.getInt(index);
    }
 
    public void putByte(int index,byte value)
    {
        this.driveFAT.put(index,value);
    }
    
    public void putShort(int index,short value)
    {
        this.driveFAT.putShort(index,value);
    }
      
    public void putInt(int index,int value)
    {
        this.driveFAT.putInt(index,value);
    }
    public void putBuffer(java.nio.ByteBuffer buf)
    {
        this.driveFAT.put(buf);
    }
  
 /*   
    public void limit(int index)
    {
        this.driveFAT.limit(index);
    }  
*/
    public void limit()
    {
        this.driveFAT.limit(this.driveFAT.capacity());
    }
     public void position(int index)
    {
        this.driveFAT.position(index);
    }  
     
     public java.nio.ByteBuffer getBuffer()
     {
         return this.driveFAT;
     }
     
     public int getClustersize()
     {
         return this.clusterSize;
     } 
     
     public java.util.ArrayList <Integer> getClusterChain()
     {
         return this.clusterChain;
     }
}
