package screens;

import smsqmulator.Monitor;
import smsqmulator.Types;

/**
 * The emulated machine's screen - a QL mode 4 or 8 screen.
 * 
 * 
 * In ql screen mode 4/0, colours for 8 pixels are in 1 word - 4 pixels per byte.
 * The lower (even) address = green bits, the higer (odd) byte of the word = red bits
 * <p> "Drawing" on the screen is done by copying the image back to the screen.
 * The "video ram" data structure is an array of short : in QL modes 4 and 8, pixels are addressed as words (8 or 4 pixels per word)
 * 
 * @author and copyright (c) 2012 -2019 Wolfgang Lenerz
 * @version
 * 1.12 clearScreen removed (→ fillscreen(0)). Better handling of display region.
 * 1.11 readXXXXFromScreen and alphaBlock removed, copyScreen moved to Screen object.
 * 1.10 copyScreen implemented.
 * 1.09 set stopAddress correctly, don't write beyond screen.
 * 1.08 implement getScreenSizeInBytes, setVibrantColours takes vram parameter, isQLScreen() introduced.
 * 1.07 better handling of edge condition (i.e. write at very last pixel on screen).
 * 1.06 slightly brighter colours possibility (not really) added - has no effect here. Part of screen initializing moved to super class.
 * 1.05 first split version. Fixed mode 8. When the object is created, the screen is set to mode 0, not mode 8.
 * (note versions up to 1.04 are for combined screens, see the Screen object).
 */
public class Screen0 extends Screen
{
    private int nbrOfPixelsPerByte=4;                     // number of pixels per byte IN MODE 4 
    private boolean isQLScreen=true;                      // are we emulating a standard QL screen?
    
    /**
     * Creates the object.
     * @param xsize the xsize of the screen in pixels
     * @param ysize the ysize of the screen in pixels. If they are both - 1, then use a QL screen.
     * @param vrambase the start of the "videoram" as far as the CPU is concerned. 
     * @param monitor the monitor running the cpu.
     * @param isMac true if we're running on a mac.
     * @param ddouble       true if double sized pixels
     * @param windowMode    mode of window 0 = window , 1  = full size,  2 = special fullsize
     * @param pctype        true if standard home etc keys
     */
    public Screen0 (int xsize,int ysize,int vrambase,Monitor monitor,boolean isMac, boolean ddouble, int windowMode, boolean pctype)
    {
        super(xsize,ysize,monitor,isMac,ddouble, windowMode, pctype);
        this.colourMode=0;
        if (xsize==512 && ysize==256)                           // emulate a standard QL screen at $20000  
        {
            this.startAddress=0x20000;
            this.stopAddress=0x28000-2;
            this.nbrOfBytesPerLine=128;                         // number of bytes per line IN MODES 4 & 8;
        }
        else
        {
            this.isQLScreen=false; 
            this.nbrOfBytesPerLine=xsize/this.nbrOfPixelsPerByte;
            this.startAddress=vrambase;
            this.stopAddress=vrambase+this.getScreenSizeInBytes(); // where screen stops (last writable word address)
        }
        fillScreen(0);          
    }
    
     /**
     * Writes a byte to the screen, painting the corresponding pixel(s).
     * 
     * @param address where to write to. This is the SMSQE address, where it thinks video memory lies.
     * It has already been checked that the address is in my memory space.
     * @param value the value to write.
     * @param wordValue the value to write into the word this byte is part of
     */
    @Override
    public void writeByteToScreen(int address, int value, int wordValue)
    {
        writeWordToScreen (address&0xfffffffe,wordValue);
    }
    
    /**
     * Writes a word to the screen, painting the corresponding pixel(s).
     * 
     * @param addr the address where to write to. It is presumed that it has been checked that this is in screen mem.
     * @param value the value to write to screen mem.
     */
    @Override
    public void writeWordToScreen(int addr, int value)
    {   
        if (addr>this.stopAddress)
            return;                                             // only useful if in QL compatible mode at address $20000
       
        // in the QL screen, the screen memory is organized in words - we have to find the word in question from the address
        value&=0xffff;
        addr-=this.startAddress;                                // where in my buffer the address lies : address is relative to start of screen
        int y=addr/this.nbrOfBytesPerLine;                      // the line (y coord) we're on
        int x=addr-(y*this.nbrOfBytesPerLine);                  // the initial column (x coord) of the pixel
        x*=this.nbrOfPixelsPerByte;                             // these are the x & y coordinates of the pixel on the screen
        y*=this.xsize;
        int colour,col;  
       
        switch (this.colourMode)
        {
            case 0:                                             // mode 4 : 1 word = 8 pixels
                x+=7;
                for (int i=0; i<8;i++)                     
                {
                    colour=(value & this.pixelmask4[i])>>>i;    // blot out unwanted bits and shift to right for each pixel
                    switch(colour)
                    {
                        case 0 :
                            col=java.awt.Color.black.getRGB();
                            break;
                        case 1 :
                            col=java.awt.Color.red.getRGB();
                            break;
                        case 0x100 :
                            col=java.awt.Color.green.getRGB();
                            break;
                        case 0x101 :
                            col=java.awt.Color.white.getRGB();
                            break;
                        default :
                            col=java.awt.Color.blue.getRGB();                      // error!
                            break;
                    }
                    this.dataBuffer.setElem(y+ x--, col);       // set in image buffer
                }
                this.isDirty=true;                              // screen needs to be redrawn
                break;
                
            case 8:                                             // mode 8 : 1 word = 4 pixels
                x+=3;
                for (int i=0; i<4;i++)                      
                {
                    colour=(value & this.pixelmask8[i])>>>(i*2);
                    switch(colour)
                    {
                        case 0 :
                            col=java.awt.Color.black.getRGB();
                            break;
                        case 1 :
                            col=java.awt.Color.blue.getRGB();
                            break;
                        case 2 :
                            col=java.awt.Color.red.getRGB();
                            break;
                        case 3 :
                            col=java.awt.Color.magenta.getRGB();
                            break;
                        case 0x200:
                            col=java.awt.Color.green.getRGB();
                            break;
                        case 0x201:
                            col=java.awt.Color.cyan.getRGB();
                            break;
                        case 0x202:
                            col=java.awt.Color.yellow.getRGB();
                            break;
                        case 0x203:
                            col=java.awt.Color.white.getRGB();
                            break;
                        default :
                            col=java.awt.Color.orange.getRGB();
                            break;
                    }
                    if (x*2+1<=this.xsize)
                    {
                        this.dataBuffer.setElem(x*2+y, col);
                        this.dataBuffer.setElem(x*2+y+1, col);
                    }
                    --x;
                }
                this.isDirty=true;
                break;
        }
    }
      
    /**
     * Writes a long word to the java screen.
     * 
     * @param addr where to write to
     * @param value the value to write
     */
    @Override
    public void writeLongToScreen(int addr,int value)
    {
        writeWordToScreen(addr,value>>>16);
        writeWordToScreen(addr+2,value);
    }

    /**
     * Sets the screen mode when in QL compatible mode.
     * Only modes 4 and 8 are supported.
     * 
     * @param mode 0 for mode 4 ,8 for mode 8
     */
    @Override
    public void setColourMode(int mode)
    {
        if ((mode!= 0 && mode !=8 ) || mode==this.colourMode)
            return;                                             // can't set this mode, or it's already set
        this.colourMode=mode;
        this.nbrOfPixelsPerByte=(mode==0)?4:2;
    }
  
    /* -------------------------------------- Component painting ------------------------------------------*/
   
   
    /**
     * Fills a block with colour - not implemented here.
     * 
     * @param cpu the Cpu used.
     * @param resolveStipple ignored.
     */
    @Override
    public void fillBlock (CPU.MC68000Cpu cpu,boolean resolveStipple)
    {
        cpu.data_regs[0]=Types.ERR_NIMP;
        cpu.reg_sr&=~4;
    }
    
    /**
     * XORS a block with colour - not implemented here.
     * 
     * @param cpu the CPu used.
     * @param resolveStrip  ignored
     */
    @Override
    public void xorBlock (CPU.MC68000Cpu cpu, boolean resolveStrip)
    {    
        cpu.data_regs[0]=Types.ERR_ORNG;
        cpu.reg_sr&=~4;
    }
    
    /**
     * This combines two blocks with alpha blending - not implemented here.
     * Not used here.
     * @param cpu the CPU used
     */
    @Override
    public  void combineBlocks(CPU.MC68000Cpu cpu)
    {
        cpu.data_regs[0]=Types.ERR_NIMP;
        cpu.reg_sr&=~4;
    }
    
    /**
     * This is called when a block of memory was copied to the vram  - not implemented here.
     * It must copy the data from the vram into the image.
     * 
     * @param cpu the CPU used.
     * @param copyFromScreen = true if the source is also the screen.
      */
    @Override
    public void moveBlock(CPU.MC68000Cpu cpu,boolean copyFromScreen)
    {   
        cpu.data_regs[0]=Types.ERR_NIMP;
        cpu.reg_sr&=~4;
    }
    
    /**
     * Displays a region if bytes were loaded directly to the screen memory.
     * Used from CPU.
     * 
     * @param cpu the cpu used.
     * @param start where the bytes were loaded to.
     * @param nbrOfBytes how many bytes were loaded.
     */
    @Override
    public void displayRegion(CPU.MC68000Cpu cpu,int start,int nbrOfBytes)
    {
        // first calculate the real start/stop adresses, making sure that we don't overstep the screen vram.
        if (this.stopAddress<start+nbrOfBytes)              // would be shoot over the top?
            nbrOfBytes-=start+nbrOfBytes-this.stopAddress;  // yes, don't
        if (start<this.startAddress)                        // would we start before the start address,?
        {
            nbrOfBytes-=(this.startAddress-start);  
            start=this.startAddress;                        // yes, don't
        }
        if (nbrOfBytes<1)                                   // now that **shouldn't** happen at all.
            return;
        // when bytes are loaded like this into screen mem, they don't necessarily make up a rectangular block.
        start&=0xfffffffe;                                  // make sure address is even=
        char []vramBuffer=cpu.getMemory();
        for (int i=0;i<nbrOfBytes;i+=2)
        {
            paintPixels(start+i,vramBuffer);
        }
        this.isDirty=true;
    }
    
    /**
     * Sets a word full of pixels in the screen image at a certain address from the vrambfuffer, but DOES NOT repaint the screen afterwards.
     * 
     * @param addr where to take the screen data from.
     */
    private void paintPixels(int addr,char[]vramBuffer)
    {
        int index=addr/2;                                   // indax into the array
        char value =vramBuffer[index];                     // get the value from the screen mem
        addr-=this.startAddress;
       
        int y=addr/this.nbrOfBytesPerLine;                  // the line (y coord) we're on
        int x=addr-(y*this.nbrOfBytesPerLine);              // the initial col (y coord) of the pixel
        x*=this.nbrOfPixelsPerByte;                         // these are the x & y coordinates of the pixel on the screen
        setPixels (x,y,value,this.colourMode);
    }     
      
    /**
     * Sets the new Vrambase for the screen buffer if we are in QL compatible mode (this implies that the CPU or its memory changed).
     * Adjust vramtop accordingly.
     * 
     * @param vrambase the new video ram base.
     */
    @Override
    public synchronized void setVramBase(int vrambase)
    {
        if (!this.isQLScreen)
        {
            this.startAddress=vrambase;
            this.stopAddress=vrambase+getScreenSizeInBytes();
        }
    }
    
    /**
     * Sets whether brighter colours should be used - ignored.
     * This does nothing here.
     * 
     * @param b <code>true</code> if brighter colours should be used.
     * @param s ignored
     */
    @Override
    public void setVibrantColours(boolean b,char[] s)
    {      
    }
    
     /**
     * Gets size, in bytes, of the vram.
     * 
     * @return screen size, in number of bytes.
     */
    @Override
    public final int getScreenSizeInBytes()
    {
        return (this.xsize*this.ysize)/this.nbrOfPixelsPerByte;
    }
    
    /**
     * Checks whether this is a QL type screen
     * 
     * @return true if this is a QL screen (512*256 i mode 4, 256*256 in mode 8), else false
     */
    @Override
    public boolean isQLScreen()
    {
        return this.isQLScreen;
    }
}
