package smsqmulator;

/**
 * The "monitor" object which, controls / starts / stops the emulation thread.
 * This is a central piece of SMSQmulator. It has a double function :
 * 
 * 1 - it serves as a monitor/debugger for M68K machine code instructions.
 * 2 - It is this monitor that launches the thread executing the instructions and also serves as a go-between between different objects.
 * 
 * <p> This executes commands that are given to it. Somme commands are:
 * <ul>
 * <li>t = trace one instruction (also enter)
 *  <li>j : jump over next instruction (don't execute it).
 *  <li>g [number_of_Instructions] : execute the program, optionally for [number_of_Instructions]
 *  <li>b [address_of_breakpoint] show (no parameter) or set/unset (one parameter) a breakpoint.
 *  <li>d [address] : display memory, if no parameter is given start at current PC address
 *  <li>di [address] : disassemble instructions, if no parameter is given start at current PC address.
 * </ul>
 * and many more, see the showHelp() method.
 * <p>
 * Many of the methods of this object are just go-between between the <code>Monitor</code> object and other objects, such as the <code>MC68000Cpu</code>object.
 * This means that some methods are called from the emulation thread, whereas others are called from the Swing Event Dispatch thread.
 *
 * The monitor outputs to two windows (instruction and data windows) and has an input wdw, see the MonitorPanel object.
 * @see smsqmulator.Monitor#showHelp() 
 * @author and copyright (c) Wolfgang Lenerz 2012-2023. Very loosely based on Tony Headford's work, see his licence below.
 * @version 
 * 
 * 1.25 large revamp, eliminated many go-between methods; there are 2 kinds of emulation thread, reset calls endEmulatiion in CPU instead of stopping the thread
 *      directly, check for success when loading SMSQE
 * 1.24 the 4 fingered salute also reloads the SMSQE file (cold boot).
 * 1.23 set screen to black as soon as reset is called.
 * 1.22 No longer passes IPHandler to TrapDispatcher, this is create ddirectly in there.
 * 1.21 added removeAllKeyrows, changes for HDD device, inputKey now directly delegates to TrapDispatcher
 * 1.20 simplified creation parameter list, also for TrapDispatcher, QL50HzInterrupt and MonitorEmulationThread, removed superfluous code and all references to throttle.
 * 1.19 use length of CPU pcs array to display content of it with "dp" command.
 * 1.18 inputMouseWheel created, when setting device names handle "./".
 * 1.17 setCopyScreen amended to suit jva_qlscremu ; setNamesForDrives; forceRemoval parameter to force "unmount" of 
 *      exiting drives and remount.
 * 1.16 setCopyScreen implemented, changeMemSize: provide for alternate CPU, reset cpu for floppy if cpu changed in changeMemSize.
 * 1.15 changeMemSize also called when new screen made, reset no longer changes screen.
 * 1.14 setScreenUpdateInterval implemented
 * 1.13 better char display in monitor itself
 * 1.12 use StringBuilder whenever appropriate
 * 1.11 call TrapDispatcher.reset() when reset.
 * 1.10 "g"o commmnd sets cpu.stopNow to 0 before executing.
 * 1.09 handle floppy disks separately ; device drivers are only loaded if they aren't disabled.
 * 1.08 the monitor no longer starts the emulation when loading the rom file, each of the guis does ;
 *      setLogger added (may set the loggers after creation).
 * 1.07 provides for be(1 - 9 = normal to fast)ep volume setting.
 * 1.06 only input mouse and keys if rom loaded ok.
 * 1.05 pass screen position to TrapDispatcher when wdw is moved/resized. needs new SMSQE.
 * 1.04 changeMemSize makes a new interrupt server (when did that get lost?).
 * 1.03 caps lock status is written directly into system variables (!).
 * 1.02 Use timezone time adjustment and user configured additional time adjustment.
 * 1.01 "Wake" Emulation thread if key typed.
 * 1.00 Throttle value is passed to this object, passed on to TrapDispatcher.
 * 0.10
 */
public class Monitor // implements java.io.Serializable
{
    private CPU.MC68000Cpu cpu;                               // the cpu we're monitoring
    private EmulationThread emulationThread=null;               // the thread that does the actual emulation
    private TrapDispatcher trapDispatcher;                      // where to dispatch calls to Java from the EmulationThread.
    private ui.MonitorGui gui;
    private final inifile.IniFile inifile;                      // file with default/configured values
    
    private boolean watchBreakpoints;                           // flag whether we observe breakpoints
    private boolean logInstructions;                            // and whether we log the instructions traced (will make everything very slow)
    private static final String COMMAND_NAMES="tgjbddikgbrhqrasxcbcdi1di2d1d2mwc1c2drdr1dr2sculslck0ck1wmspewrdp";// what commands are recognized by the monotir
    private StringBuilder sbuffer;                              // avoid too much string creation
    private static final int ERRORCODE=-123456789;              // an error code I use
    private java.util.ArrayList<Integer> breakpoints=new java.util.ArrayList<>(); //list of breakpoints
    private String debugFilename;                               // where to put the logged instructions
    private int diPanelNbr=1;                                   // display instructions panel
    private int dPanelNbr=2;                                    // data panel
    private javax.swing.JTextArea dTextArea;                    // the data window (to the left)
    private javax.swing.JTextArea diTextArea;                   // the instruction window (to the right)
    private javax.swing.JTextField inputWdw;                    // the input wdw, where you enter commands
    private javax.swing.JTextArea regLogger;                    // where we display
    private javax.swing.JTextArea dataLogger;               
    private String lastCommand="";                              // last monitor command given
    private int conditionDataReg=-1;                            // these ...
    private int conditionAddressReg=-1;
    private int conditionAddress=-1;
    private int conditionValue;
    private boolean conditionIsContent;                         // ... determine when emulation stops at a condition
    private int upperLimit=0;                                   // no mem checked above that
    private boolean checkmem=false;                             // do we check memory?
    private int memoryToBeWatched=0;                            // memory location possibly watched
    private boolean excludeSuper=false;                         // do we exclude calls to supervisor mode when debugging?
    private static final int MEM_FOR_ROM=350000;                // mem needed to load SMSQE. Increase if the SMSQE file size exceeeds this
    private final int nbrLogs;
    
    /**
     * Creates the object
     * 
     * @param watchBreakpoints flag whether we observe breakpoints (<code>true</code> = yes).
     * @param inifile the ".ini" file with initial and current values.
     * @param nbrLogs whether we log the instructions traced (will make everything very slow), and how many. =0 = don't log
     */
    public Monitor (boolean watchBreakpoints,inifile.IniFile inifile, int nbrLogs,boolean logAtStartup)    
    {
        this.watchBreakpoints=watchBreakpoints;
        this.sbuffer = new StringBuilder(240);
        this.inifile=inifile;
        this.nbrLogs=nbrLogs;
        this.logInstructions=logAtStartup;
        try
        {
            this.debugFilename=(System.getProperty( "user.home" )+java.io.File.separator+"SMSQmulatorDebug.txt");// log file
        }
        catch (Exception e)
        {/*nop*/}                                               // if I can't get the debug log file, then that's not the end of the world                   
        
        if (this.regLogger==null)                               // make sure this exists
            this.regLogger=this.dataLogger;
        if (this.regLogger==null)
        {
            this.regLogger=new javax.swing.JTextArea();         // if it still doesn't create both
            this.dataLogger=this.regLogger;
        }
    }

    /**
     * Sets the gui for this monitor.
     * 
     * @param gui 
     * @param trapDispatcher 
     */
    public void setGuiAndCpu(ui.MonitorGui gui,TrapDispatcher trapDispatcher)
    {
        this.gui=gui;
        this.regLogger=gui.getRegLogger();
        this.dataLogger=gui.getDataLogger();
        this.inputWdw=gui.getInputWindow();
        this.diTextArea=getPanel(this.diPanelNbr);
        this.dTextArea=getPanel(this.dPanelNbr);
        this.trapDispatcher=trapDispatcher;
        this.trapDispatcher.setGui(gui);
    }
    
    /*-------------------------------------------------------------- Monitor commands themselves ---------------------------------------------------------*/
    
    /**
     * Executes a command.
     * <p> See the help for what commands are recognized by the monitor.
     * 
     * @param command a command line : command [options]:
     * <p>Commands are single or double letters, possibly followed by options.
     * <p>Some commands are (there are more,see the help method):
     * <ul>
     * <li>   h - Help: show help.
     * <li>   g - Go (start monitoring) &lt;start_address (default current PC)&gt; &lt;number_of_instructions to go (default unlimited)&gt;.
     * <li>   gb- Go to a provisional breakpoint. One parameter: the breakpoint.
     * <li>   r - Registers : show content of registers.
     * <li>   a - check whether the go thread is still running.
     * <li>   di- Disassemble instructions &lt;start_address (default current PC)&gt; &lt;number_of_instructions (default : 8)&gt;.
     * <li>   j - Jump over current instruction (do not execute, advance PC).
     * <li>   b - Breakpoint. No parameters = sh;ow breakpoints , parameter= set breakpoint at that addresss, or unset it if it already is set at that address.
     * <li>   t - Trace (execute) one instruction at the current PC.
     * <li>   d - display memory content &lt;start_address (default current PC)&gt; &lt;number_of_bytes (default : 16)&gt;.
     * <li>   k - kill current Go thread if it exists.
     * <li>   s - set a register to a value: s &lt;reg&gt; &lt;value&gt; or s &lt;reg&gt; &lt;value&gt;.
     * <li>   bc - clear all breakpoints.
     * <li>   qr- quickly execute the following instruction and set a breakpoint to the one after that. Useful for JSR/BSR/Trap.
     * </ul>
     * <p> Options are separated from the command and each other by single spaces.
     * <p> Numbers can be given in decimal or hex format. The default is hex, decimals must be prefixed by "&amp;".
     * <p> An address can also be given by register number enclosed in parenthesis, eg (a7) or (d0) and also (pc).
     * <p> The address will then be that of the content of the register, so, if a7= $100, 'di (a7)' will disassemble instructions as of address $100.
     * <p> An address can also be calculated with + or - , e.g. *-10 or (a0)+$100.
     * @see smsqmulator.Monitor#showHelp() 
     */
    public void executeMonitorCommand (String command)
    {
        if (command==null)
            return;                                             // no nonsense, please
        command=command.toLowerCase().trim();                 
        if (command.isEmpty())
            command=this.lastCommand;                           // simple enter = do last command again
        if (command.isEmpty())
            return;                                             // if no last command, just do nothing
        int t=command.indexOf("  ");                            // any double spaces?
        while (t!=-1)
        {
            command=command.replaceAll("  "," ");               // make sure there are no double spaces at all
            t=command.indexOf("  ");
        }
        String [] commands=command.split(" ");
        if (commands.length==0)
            return;                                             // how could that be?
        int commandIndex=Monitor.COMMAND_NAMES.indexOf(commands[0]);// what command is it?
        if (commandIndex==-1 && command.startsWith("s"))
            commandIndex=15;                                    // special case Sd0-Sd7 and Sa0-Sa7
        switch (commandIndex)
        {
            case 0:                                             // "t"race
                this.cpu.execute();
                this.cpu.stopNow=0;
                showInfo(true,this.diTextArea);
                this.lastCommand=command;
                break;
            case 1:                                             // "g"o
                cpu.stopNow=0;
                goCommand(commands,-1);
                this.lastCommand="";
                break;
            case 2:                                             // "j"ump over the next command
                int address=this.cpu.pc_reg*2;
                CPU.DisassembledInstruction di = getDI(address);
                address+=di.size();                             // total size of instruction at that address
                this.cpu.pc_reg=address/2;                      // set PC to that address
                showInfo(true,this.diTextArea);                 // show new state
                this.lastCommand="";
                break;
            case 3:                                             // "b"reakpoint (show, set or delete)
                setBreakpoints(commands);
                this.lastCommand="";
                break;
            case 4:                                             // "d"isplay mem
                dispMem(commands);
                this.lastCommand=command;
                break;
            case 5:                                             // "di" display instructions
                diCommand(commands);
                this.lastCommand=command;
                break;
            case 7:                                             // "k"ill go thread
                killEmulation();                                // if there is a thread, tell it to stop
                this.lastCommand="";
                break;
            case 8:                                             // "GB" - go till breakpoint (one option : the breakpoint)
                if (commands.length>1)
                {
                    address=parseForInteger(commands[1]);
                    if (address!=Monitor.ERRORCODE)
                    {
                        goCommand(null,address);
                    }
                }
                else
                {
                    showError ("Error - gb needs one parmeter","qb <breakpoint address>",null);
                }
                this.lastCommand="";
                break;
            case 10:                                            // "R" : show registers
            case 38:                                            // "DR1": set wdw
                this.diTextArea=this.dataLogger;
            case 36:                                            // "DR"
                showInfo(true,this.diTextArea);
                this.lastCommand="";
                break;
            case 11:                                            // "H"elp
                showHelp();
                this.lastCommand="";
                break;
            case 12:                                            // "qr" quick trace over.
                traceOver();
                this.lastCommand="t";                           // and pretend last command was a trace
                break;
            case 14:                                            // "A" -show status of gothread
                if (this.emulationThread!=null)
                    this.regLogger.append("Emulation thread still running...!\n");
                else
                    this.regLogger.append("Emulation thread is stopped.\n");
                this.lastCommand="";
                break;
            case 15:                                            // "s" set a register with a value
                setRegister (commands);
                this.lastCommand="";
                break;
            case 16:                                            // X just execute the emulation as fast as possible (i.e. go into fast mode)
              //  this.lastCommand="";                          // does nothing
                break;
            case 17:                                            // "c" clear debugger windows
                this.dataLogger.setText("");
                this.regLogger.setText("");
                break;
            case 18:                                            // "BC" clear breakpoints
                this.breakpoints.clear();
                this.lastCommand="";
                this.regLogger.setText("All breakpoints cleared");
                break;
            case 20:
            case 23:                                            // di1 or di2 display instructions in wdw 1 or 2
                dixCommand(commands);
                this.lastCommand=command;
                break;
            case 26:
            case 28:                                            // d1 or d2 display data in wdw 1 or 2
                dispXMem(commands);
                this.lastCommand=command;
                break;
            case 30:                                            // "mw" set content of memory in words
		setMem(commands);
                break;
            case 32:                                            // "c1" clear wdw 1 only
                this.dataLogger.setText("");
                break;
            case 34:                                            // "c2" clear wdw 2 only
                this.regLogger.setText("");
                break;
            case 41:                                            // "dr2" display registers in wdw 2
                showInfo(true,this.regLogger);
                this.lastCommand="";
                break;
            case 44:                                            // "SC" set condition
                parseCondition (commands);
                break;
            case 46:                                            // "UL" set upper limit of mem for logging
                setUpperLimit(commands);
                break;
            case 48:                                            //"SL" switch logging on/off
                this.logInstructions=!this.logInstructions;
                this.regLogger.setText("Logging is now "+ (this.logInstructions? "ON" : "OFF")+"\n");
                if (this.emulationThread!=null)
                    this.emulationThread.switchLogging(this.logInstructions);
                break;
            case 50:                                            // ck0, ck1 = stop/start checking memory to be watched
            case 53:
                this.checkmem=commandIndex==53;
                if (this.emulationThread!=null)
                    this.emulationThread.setckmem(this.checkmem);
                break;
            case 56:                                            // "wm" set memory to be watched (watching is started/stopped with ck1 and ck2 above!)
                setMemToBeWatched(commands);
                break;
            case 58:                                            // "sp" special - does nothing any more, see bits of code after end of object
            case 60:                                            // e - exclude rom calls when tracing DO NOTHING ANYMORE (called excludeROM)
                break;                                          
            case 61:                                            // wr write NIY
              //  writeFile();
                break;                                         
            case 63:                                            // dp display last pc values
                this.regLogger.setText("");
                for (int k=0;k<this.cpu.pcs.length;k++)
                {
                    this.regLogger.append(String.format("%08x", this.cpu.pcs[k]*2).toUpperCase()+"\n");
                }
        }
    }
    
    /**
     * Sets the memory address that is to be watched for changes.
     * Watching is switched on/off with the ck1 and ck0 commands.
     * 
     * @param options  element 1 contains the memory address that is to be watched for changes.
     */
    private void setMemToBeWatched(String []options)
    {
        if (options.length!=2)
        {
            showError ("wm - wrong ","wm <address>",options);
            return;
        } 
        if (options[1].equals("?"))
        { 
            String s;
            if (this.memoryToBeWatched<0)
                s= "NO mem is being watched";
            else                
                s= "Mem watch set at $"+Integer.toHexString(this.memoryToBeWatched) + " - Current value is "+this.cpu.readMemoryLong(this.memoryToBeWatched);
            s+="\n Mem watching is switched "+(this.checkmem?"ON":"OFF")+"\n ";
            
            this.regLogger.setText(s);
            return;
        }
        int address=parseForInteger(options[1]);
        if (address==Monitor.ERRORCODE)
        {
            showError ("wm - wrong ","wm <address>",options);
            return;
        }
        this.memoryToBeWatched=address; 
        if (this.emulationThread!=null)
            this.emulationThread.setmemWatched(address);
        
        this.regLogger.setText("Mem watch set at "+String.format("%08x", address) + " - Current value is "+this.cpu.readMemoryLong(address)+ "\n");
    }
    
    /**
     * Sets the upper limit of memory when logging.
     * 
     * @param options  element 1 contains the memory address. When logging, code executed above this is not logged.
     */
    private void setUpperLimit(String []options)
    {
        if (options.length!=2)
        {
            this.regLogger.setText("Upper limit = "+String.format("%08x",this.upperLimit)+"\n");
            return;
        } 
        int address=parseForInteger(options[1]);
        if (address==Monitor.ERRORCODE)
        {
            showError ("ul - wrong ","ul <address>",options);
            return;
        }
        this.upperLimit=address;
        if (this.emulationThread!=null)
            this.emulationThread.setUpperLimit(address);
    }
   
    /**
     * Sets content of memory, always in words.
     * 
     * @param options  element 1 contains the memory address, element 2 contains the new memory content at that address.
     */
    private void setMem(String []options)
    {
        if (options.length!=3)
        {
            showError ("mw - wrong number of instructions","mw <address> <number>",options);
            return;
        }
        int address=parseForInteger(options[1]);
        int mem=parseForInteger(options[2]);
        if (address==Monitor.ERRORCODE || mem ==Monitor.ERRORCODE)
        {
            showError ("mw - wrong number of parameters","mw <address> <number>",options);
            return;
        }
        this.cpu.writeMemoryWord(address, mem&0x0000ffff);
        showInfo(true,this.diTextArea);
    }
    
    /**
     * QR = Traces over the instruction until the next instruction.
     * 
     * <p> If the instruction is a jump, this will trace until the rts.
     */
    private void traceOver()
    {
        int address=this.cpu.pc_reg*2;
        CPU.DisassembledInstruction di = getDI(address);// get instruction (and its size)
        int stop=address+di.size();
        while (address!=stop)
        {
            int ret=this.cpu.execute();
            if (ret<-1)
                 this.trapDispatcher.dispatchTrap(ret,this.cpu);
            address=this.cpu.pc_reg*2;
        }
        showInfo(true,this.diTextArea);
    }
    
    /**
     * Handles the "go" command : execute all instructions until either breakpoint reached or the optional nbr of instructions is reached.
     * <p>The MonitorEmulationThread handles actual instuction execution, and can thus be stopped with the "k" commmand.
     * 
     * @param options Options[1] optionally contains the number of instructions to execute.
     * @param provbkp a provisional break for the "GB" command.
     */
    public void goCommand(String []options,int provbkp)
    {
        if (this.emulationThread!=null)                         // don't execute the Go thread twice!
        {
            this.regLogger.setText("Error: Emulation already or still running (go command? - may be killed with k)");
            return;                                             // PREMATURE EXIT
        }
        int nbrInst=0;
        if (options!= null && options.length>1)                 // any options?
        {
            nbrInst=parseForInteger(options[1]);                // yes, try to get the number of instructions
            if (nbrInst==Monitor.ERRORCODE)
            {
                showError ("g - wrong number of instructions","g [<nbr of instructions>]",options);
                return;                                         // PREMATURE EXIT
            }
        }
        this.emulationThread=new MonitoredEmulationThread(nbrInst,this,this.watchBreakpoints,this.upperLimit,this.checkmem,
                                                          this.memoryToBeWatched,provbkp,this.excludeSuper,this.gui.getScreen(),
                                                          this.logInstructions,this.nbrLogs);
        setCondition();
        this.emulationThread.start();
        if (this.gui!=null)
            this.gui.setFocus();
    }    
    
    /**
     * di1 or di2 : Disassembles instructions in the text area given by the command.
     * @param options    Options[1] optionally contains the address where to start disassembling (default = PC),
     * <p>               Options[2] optionally contains the number of instructions to disassemble (default=16).
     */
    private void dixCommand(String []options)
    {
        int panelNbr=1;                                         // default display panel
        try
        {
            panelNbr=Integer.parseInt((options[0].substring(2)));// try to get panel nbr
        }
        catch (NumberFormatException e)
        { /*NOP*/ }                                             // keep default if pb
        this.diTextArea=getPanel(panelNbr);                     // set textarea to the one wished
        diCommand(options);                                     // show instructions
    }

    /**
     * Disassembles Instructions.
     * @param options   Options[1] optionally contains the address where to start disassembling (default = PC),
     * <p>                 Options[2] optionally contains the number of instructions to disassemble (default=16).
     */
    private void diCommand(String []options)
    {
        int nbrInst=16;
        int address=this.cpu.pc_reg*2;
        if (options.length>1)
        {
            address=parseForInteger(options[1]);                // try to get the start address
            if (address==Monitor.ERRORCODE || address<0)
            {
                showError ("di - wrong start address","di [<start_address> [<nbr of bytes>]]",options);
                return;
            }
        }
        if (options.length>2)
        {
            nbrInst=parseForInteger(options[2]);                // try to get the number of instructions
            if (nbrInst==Monitor.ERRORCODE)
            {
                showError ("di - wrong number of bytes","di [<start_address> [<nbr of bytes>]]",options);
                return;
            }
        }
        StringBuilder sb = new StringBuilder(80);
        int memsize=this.cpu.readableSize();
        for (int i=0;i<nbrInst && address<memsize;i++)
        {
            sb.delete(0, 80);
            CPU.DisassembledInstruction di = getDI (address);
            di.shortFormat(sb);//di.formatInstruction(buffer);
            this.diTextArea.append(sb.toString().toUpperCase()+"\n");
            address += di.size();
        }
    }
    
    /**
     * d1 or d2 : Displays memory in the text area given by the command.
     * @param options Options[1] optionally contains the address where to start showing mem (default = PC)
     *                Options[2] optionally contains the number of bytes to show (default 256).
     */
    private void dispXMem(String []options)
    { 
        int panelNbr=1;
        try
        {
            panelNbr=Integer.parseInt((options[0].substring(1)));// get panel nbr
        }
        catch (NumberFormatException e)
        { /*NOP*/ }                                         // keep default 1
        this.dTextArea=getPanel(panelNbr);                 // set textarea to the one wished
        dispMem(options);  
    }

    /**
     * Displays memory.
     * 
     * @param options Options[1] optionally contains the address where to start showing mem (default = PC)
     *                Options[2] optionally contains the number of bytes to show (default 256).
     */
    private void dispMem(String []options)
    {
        int nbrBytes=16 *16;                                    // 16 ROWS @ 16 BYTES        
        int address=this.cpu.pc_reg*2;
        if (options.length>1)
        {
            address=parseForInteger(options[1]);                // try to get the start address
            if (address==Monitor.ERRORCODE || address<0)
            {
                showError ("d - wrong start address","d [<start_address> [<nbr of bytes>]]",options);
                return;
            }
        }
               
        if (options.length>2)
        {
            nbrBytes=parseForInteger(options[2]);               // try to get the number of bytes
            if (nbrBytes==Monitor.ERRORCODE)
            {
                showError ("d - wrong number of bytes","d [<start_address> [<nbr of bytes>]]",options);
                return;
            }
        }
        
        StringBuilder sb = new StringBuilder(80);               // the bytes as hex
        StringBuilder p = new StringBuilder(80);                // the bytes as chars
        int end=address+nbrBytes;
        int b;
        char c;
        while (address<end)
        {
            sb.setLength(0);
            sb.append(String.format("%08x", address).toUpperCase()).append(" ");
            p.setLength(0);
            p.append("  ");
            for (int i=0;i<8 && address<end; i++)
            {
                if (i==4)
                    sb.append(" ");
                b = cpu.readMemoryByte(address++);
                sb.append(String.format("%02x", b).toUpperCase()); // it would be faster to read words, but address may be odd
                c=(b<32 || b>126)?'.':(char)b;
                p.append(c);
                b = cpu.readMemoryByte(address++);
                sb.append(String.format("%02x", b).toUpperCase()); 
                sb.append(" "); 
                c=(b<32 || b>126)?'.':(char)b;
                p.append(c);
            }
            p.append("\n");
            this.dTextArea.append (sb.toString()+p.toString());
        }
    }
    
    /**
     * Sets a register to a value.
     * 
     * @param commands  element 0 contains command+register, element 1 contains the value OR 
     *<p>           element 0 contains command ("s"), element 1 contains register, element 2 contains the value
     */
    private void setRegister (String [] commands)
    {
        if (commands.length<2)
        {
            showError ("setRegister","s<register> <content> or s <register> <content>",commands);
            return;
        }
        String reg,value;
        if (commands[0].length()!=1)                        // is it "sxx" or "s xx"?
        {
            reg=commands[0];
            if (reg.length()!=3)                        
            {
                showError ("setRegister","s<register> <content> or s <register> <content>",commands);
                return;
            }
            reg=reg.substring(1);
            value=commands[1];
        }
        else
        {
            if (commands.length<3)
            {
                showError ("setRegister","s<register> <content> or s <register> <content>",commands);
                return;
            }
            reg=commands[1];
            value=commands[2];
        }
        // here now reg=the register, value = the value;
        int val=parseForInteger(value);
        if (val==Monitor.ERRORCODE || reg.length()!=2)
        {
            showError ("setRegister","s<register> <content> or s <register> <content>",commands);
            return;
        }
        int register="01234567".indexOf(reg.substring(1));      // reg nbr
        if (register==-1 && (!reg.equals("pc")))
        {
            showError ("setRegister","s<register> <content> or s <register> <content>",commands);
            return;
        }
        int type="adp".indexOf(reg.substring(0,1)); 
        switch (type)
        {
            case 0:
                this.cpu.addr_regs[register]= val;
                showInfo(true,this.diTextArea);
                break;
            case 1:
                this.cpu.data_regs[register]= val;
                showInfo(true,this.diTextArea);
                break;
            case 2:
                this.cpu.pc_reg=val/2;
                break;
            default:
                showError ("setRegister","s<register> <content> or s <register> <content>",commands);
                break;
        }
    }
      
    /**
     * Sets/unsets a breakpoint.
     * If a breakpoint is already set at the given address it will be unset, else it will be set.
     * 
     * @param options if only 1 element : show all breakpoints, else element 1 = address where to set/unset breakpoint
     */
    private void setBreakpoints(String []options)
    {
        if (options.length==1)                          
        {
            this.regLogger.setText("Breakpoints: ");
            for (int i: this.breakpoints)
                this.regLogger.append(String.format("%08x (%d)",i,i).toUpperCase()+"\n");
        }
        else
        {
            int address=parseForInteger(options[1]);
            if (address==Monitor.ERRORCODE)
            {
                showError("breakpoint","b [< address>]",options);
                return;
            }
           
            if (this.breakpoints.contains(address))
            {
                this.breakpoints.remove((Integer)address);
                this.regLogger.append(String.format("Breakpoint removed at $%08x\n",address).toUpperCase());
            }
            else
            {
                this.breakpoints.add(address);
                this.regLogger.append(String.format("Breakpoint set at %08x\n",address).toUpperCase());
            }
        }
    }

    /**
     * Parses the condition : sc "<condition>". SC alone resets the condition.
     * condtiion is reg =value,
     * eg a6=x..
     * d6=x..
     * OR!!! c4=x... (content of address pointed to by address reg)
     * 
     * OR x=(address) sets the address for the condition
     * 
     * where c4 then denotes that this is the content of address reg 4
     */
    private void parseCondition(String []options)
    {
        if (options.length==1)                              // "sc" alone
        { 
            this.conditionDataReg=-1;
            this.conditionAddressReg=-1;
            this.conditionValue=0;  
            this.conditionAddress=-1;
            this.conditionIsContent=false;
            setCondition();
            return;
        } 
        if (options.length!=2)                              
        {
            showError("set condition","sc <condition>",options);
            return;
        }
        String cond=options[1];                             // the coondition
        if (cond==null || cond.isEmpty())                   // must't be empty
        {
            showError("set condition","sc <condition>",options);
            return;
        }
        String []condvals=cond.split("=");
        if (condvals==null || condvals.length!=2)           // condition format is reg=value
        {
            showError("set condition","sc <condition>",options);
            return;
        }
        String condreg=condvals[0].toLowerCase().trim();    // get reg  
        int value=parseForInteger(condvals[1].trim());      // get value
        if (condreg.length()!=2 || value==Monitor.ERRORCODE)
        {
            showError("set condition","sc <condition>",options);
            return;
        }
        int regtype="adcx".indexOf(condreg.substring(0,1));      // which type of reg is it?
        if (regtype<0)
        {
            showError("set condition","sc <condition>",options);
            return;
        }
        int temp=parseForInteger(condreg.substring(1));     // reg number
        if (temp==Monitor.ERRORCODE || temp>7 || temp<0)    // must be between 0 & 7 
        {
            showError("set condition","sc <condition>",options);
            return;
        }
        switch (regtype)
        {
       
            case 0:                                             // condition is for address reg
                this.conditionDataReg=-1;
                this.conditionAddressReg=temp;
                this.conditionValue=value;
                this.conditionIsContent=false;
                setCondition();
                break;
            case 1:                                             // data reg
                this.conditionDataReg=temp;
                this.conditionAddressReg=-1;
                this.conditionValue=value;
                this.conditionIsContent=false;
                setCondition();
                break;
            case 2:                                             // content of mem pointed to fy address regs
                this.conditionDataReg=-1;
                this.conditionAddressReg=temp;
                this.conditionValue=value;
                this.conditionIsContent=true;
                setCondition();
                break;
            case 3:                                             // condition at a certain address only
                this.conditionAddress=value;
                setCondition();
                break;
            default:
                break;
        }
    }
    
    /**
     * Sets a condition (reg=value).
     */
    private void setCondition()
    {
        if (this.emulationThread!=null)
        {
            this.emulationThread.setCondition (this.conditionDataReg,this.conditionAddressReg,this.conditionValue,
                    this.conditionIsContent,this.conditionAddress);
        }
    }
    
    /**
     * Sets the log windows.
     * 
     * @param reglogger where to show the register contents.
     * @param datalogger where to show data, e.g. memory content.
     */
    public void setLoggers (javax.swing.JTextArea reglogger,javax.swing.JTextArea datalogger)
    {
        this.regLogger=reglogger;
        this.dataLogger=datalogger;
        this.diTextArea=getPanel(this.diPanelNbr);
        this.dTextArea=getPanel(this.dPanelNbr);
    }               
  /**
     * Gets the list of breakpoints.
     * @return the list of breakpoints may be empty but will not be null.
     */
    public  java.util.ArrayList<Integer> getBreakpoints()
    {
        return this.breakpoints;
    }
    
    /**
     * Get the name of the logfile.
     * 
     * @return  the name of the logfile.
     */
    public String getLogfilename()
    {
        return this.debugFilename;
    }
    
    /**
     * Shows the help.
     */
    private void showHelp()
    {
        this.dataLogger.setText("");
        this.dataLogger.append("Commands are single or double letters, possibly followed by options.\n");
        this.dataLogger.append("Commands are:\n");
        this.dataLogger.append("  h - Help: show help.\n");
        this.dataLogger.append("  g - Go (start monitoring) <start_address (default current PC)> <number_of_instructions to go (default unlimited)>.\n");
        this.dataLogger.append("  gb- Go to a provisional breakpoint. One paramter: the breakpoint.\n");
        this.dataLogger.append("  r and dr - display registers : show content of all registers (no parameter).\n");
        this.dataLogger.append("  dr1 and dr2 - display registers in window 1 or 2\n");
        this.dataLogger.append("  a - check whether the go thread is still running.\n");
        this.dataLogger.append("  di- Disasssemble instructions <start_address (default current PC)> <number_of_instructions (default : 8)>.\n");
        this.dataLogger.append("      This can also be 'di1' or 'di2' to choose the display area where the result will be shown\n");
        this.dataLogger.append("  j - Jump over current instruction (do not execute, advance PC to next instruction).\n");
        this.dataLogger.append("  b - Breakpoint. No parameters = show breakpoints , parameter= set breakpoint at that addresss, or unset it, if it already is set at that address.\n");
        this.dataLogger.append("  t - Trace (execute) one instruction at the current PC.\n");
        this.dataLogger.append("  d - display memory content <start_address (default current PC)> <number_of_bytes (default : 16)>.\n");
        this.dataLogger.append("      This can also be 'd1' or 'd2' to choose the display area where the result will be shown\n");
        this.dataLogger.append("  k - kill current Go thread if it exists.\n");
        this.dataLogger.append("  s - set a register to a value: s<reg> <value> or s <reg> <value>.\n");
        this.dataLogger.append("  bc - clear all breakpoints.\n");
        this.dataLogger.append("  qr- quickly execute the following instruction & set a breakpoint to the one after that. Useful for JSR/BSR/Trap.\n");
        this.dataLogger.append("      This does NOT stop at any normal breakpoints.\n");
        this.dataLogger.append("  mw- change a word in memory : mw <address> <new word content>.\n");
        this.dataLogger.append("  ck0, ck1 swicth watching memory on (ck1) or off (ck2). When on, and watched memory != 0 any change in the long value in that memory location will trigger a breakpoint.\n");
        this.dataLogger.append("  sc- set a condition : sc <condition> <address> sets the condition at that address. a condition is <register>=<value> WITHOUT spaces in between, eg a0=28000\n");
        this.dataLogger.append("  ul- set upper limit : when logging instructions, if the pc is beyond this limit, instructions will not be logged.\n");
     //   this.dataLogger.append("  e  - exclude ROM code from logging (switch on/off).\n");
        this.dataLogger.append("  sl - switch logging instructions from yes to no to yes etc When switched on and then off, the logged instructions are \n");
        this.dataLogger.append("       written into a log file, by default this is (user home)+SMSQmulatorDebug.txt.\n");
     //   this.dataLogger.append("  wf - write file : write the logging file immediately.\n");
        this.dataLogger.append("  dp - display the previous 10 PCs - uncomment the relevant section in the CPU.execute method!!!!\n");
        this.dataLogger.append("  wm - watched memory : wm <address> Sets the memory location that will be watched.\n");
        this.dataLogger.append("  ck1 & ck0 start/stop watching memory.\n");
      //  this.dataLogger.append("  sp - special, could be anything.\n");
        this.dataLogger.append("A simple ENTER will repeat the last t, d or di command.\n");
        this.dataLogger.append("Options are separated from the command and each other by single spaces.\n");
        this.dataLogger.append("Numbers can be given in decimal or hex format. The default is hex, decimals must be prefixed by '&'.\n");
        this.dataLogger.append("The star '*' can be used for an address, this then is the current PC \n");
        this.dataLogger.append("An address can also be given by register number enclosed in parenthesis, eg (a7) or (d0) and also (pc):\n");
        this.dataLogger.append("The address will then be that of the content of the register, so, if a7= $100, 'di (a7)' will disassemble instructions as of address $100.\n");
        this.dataLogger.append("An address can also be calculated with + or - , e.g. *-10 or (a0)+$100.\n");
        this.dataLogger.append("An address can also be calculated with a displacement eg 10(a0) and combined regs eg. 10(a3,a6). Both regs will be long.\n");
        
    }

/*---------------------------------------------------  Monitor commands subroutines -------------------------------------------------------*/
    /**
     * Shows that an error occurred in a command.
     * 
     * @param command the command with the error.
     * @param correct the correct way to use the command.
     * @param options the command line that was used instead.
     */
    private void showError(String command, String correct,String []options)
    {
        StringBuilder p=new StringBuilder();
        if (options!=null)
        {
            for (String m:options)
                p.append(m).append(" ");
        }
        this.regLogger.setText("Error in instruction -"+command+" : "+p.toString()+"\n");
        this.regLogger.append("Usage : "+correct+"\n");
    }
    
    /**
     * Parses a string for an integer.
     * 
     * @param s the string to parse as integer. This may be a hex integer or a decimal (it is then prefixed by "&", e.g. &0123),
     *        or the char "*" which then means the current PC address of the CPU,
     *        or a data or address reg in parentheses (eg :" (a7) " and also (a6,a7)).  
     *        or any of the above and a displacement eg *+45 or even (a7)+(a6);
     * 
     * @return the result. This will be <code>Monitor.ERRORCODE</code> if there was any problem with the string.
     */
    private int parseForInteger(String s)
    {
        int result=Monitor.ERRORCODE;                           // signal error
        if (s==null || s.isEmpty())
            return result;                                      // null or empty string  - oops
        s=s.trim(); 
        if (s.isEmpty())
            return result;                                      // string was only spaces - oops
        
        String additionString;
        int addedValue=0,temp;
        int t=s.indexOf("(");                                   // any index or regs?
        if (t>0)
        {
            addedValue=parseForInteger(s.substring(t));         // yes, parse again as of the "("
            if (addedValue==Monitor.ERRORCODE)
                return addedValue;
            s=s.substring(0,t);
        }
        
        boolean minus=false;
        t=s.indexOf("-");                                       // any minus sign in string?...
        if (t==-1)
            t=s.indexOf("+");                                   // ...no, any plus sign?
        else
            minus=true;                                         // ....yes, there was a minus sign
        if (t!=-1)                                              // was there a + or - sign?
        {
            if (t+1>=s.length())                                // plus or minus sign at end? error
                return result;                      
            additionString=s.substring(t+1);                    // the string, stripped of the sign
            temp=parseForInteger(additionString);               // get as address (or offset)
            if (temp==Monitor.ERRORCODE)
                return temp;      
            if (minus)
                temp*=-1;
            addedValue+=temp;
            s=s.substring(0,t);
            if (s.isEmpty())
                return addedValue;
        }
            
        if (s.equals("*"))
            return this.cpu.pc_reg*2+addedValue;
        if(s.startsWith("("))                                   // is it something like "(a7)"?
        {
            if (!(s.endsWith(")")))                             // string not enclosed in ()?
                return result;
            t=s.indexOf(",");                                   // composed registers eg. (a6,a3)?       
            if (t==-1)                          
            {                                                   // no, single reg
                if(s.length()!=4)
                    return result;                              // oops, something wrong
                String a=s.substring(1,2);                      // the register, either "a" or "d" or "p"
                if (!("adp".contains(a)))                       // but it isn't - oops
                    return result;
                try
                {
                    if (a.equals("p"))
                        return this.cpu.pc_reg*2+addedValue;    // reg was (pc)
                    result=Integer.parseInt(s.substring(2,3));  // get the register number
                    if (result<0 || result>7)
                        return Monitor.ERRORCODE;               // regs go from 0 to 7 only
                    if (a.equals("a"))
                        return this.cpu.addr_regs[result]+addedValue;// get content of addr reg
                    else
                        return this.cpu.data_regs[result]+addedValue;

                }
                catch(Exception e)
                {
                    return Monitor.ERRORCODE;                   // register number isn't
                }
            }
            else                                                // composed regs eg (a6,a1) - they will always be taken as long
            {   
                if ((t!=3) || s.length()!=7)
                    return result;                              // something wrong
                t=parseForInteger("("+s.substring(1,3)+")");    // get value of first reg
                if (t==result)                                  // oops
                    return result;
                result=parseForInteger("("+s.substring(4,6)+")");// second reg, get its value
                if (result==Monitor.ERRORCODE)
                    return result;                              // oops
                return t+result+addedValue;                     // return combined values
            }
        }
        int radix=16;                                           // signals hex number as default
        if (s.startsWith("$"))
        {
            if (s.length()==1)                                  // string contains only a "$"
                return result;
            s=s.substring(1);
            radix=16;
        }
        else if (s.startsWith("&"))
        {
            if (s.length()==1)                                  // string contains only a "&"
                return result;
            s=s.substring(1);
            radix=10;
        }
        try
        {
            boolean isNeg=false;
            if (s.length()>8 && radix==16)
                return result;                                  // hex values only accept 8 chars
            if (s.length()==8 && radix ==16)
            {
                String m="0"+s.substring(0,1);                  // get first nibble of 32 bit number
                try
                {
                    temp=Integer.parseInt(m,radix);
                }
                catch (Exception e)
                {
                    return result;
                }
                if (temp>7)
                {
                    isNeg=true;
                    temp-=8;
                    s=temp+s.substring(1);
                }
            }
            result=Integer.parseInt(s,radix);
            if (isNeg)
                result |=0x80000000;
        }
        catch (NumberFormatException e)
        {
            return Monitor.ERRORCODE;
        }
        return result+addedValue;
    }
    
    /**
     * Shows the content of all regs and the next instruction.
     * 
     * @param clearArea should the text area be cleared before info is shown?
     * @param textArea where to show the info.
     */
    public void showInfo(boolean clearArea,javax.swing.JTextArea textArea)
    {
        if (textArea==null)
            textArea=this.diTextArea;
        if(clearArea)
            textArea.setText("");
        textArea.append(String.format("D0:%08x  D1:%08x  D2:%08x  D3:%08x  PC:%08x\n",
                        this.cpu.data_regs[0], this.cpu.data_regs[1], this.cpu.data_regs[2], this.cpu.data_regs[3],this.cpu.pc_reg*2).toUpperCase());
        textArea.append(String.format("D4:%08x  D5:%08x  D6:%08x  D7:%08x  SR:%04x %s\n",
                        this.cpu.data_regs[4], this.cpu.data_regs[5], this.cpu.data_regs[6], this.cpu.data_regs[7],this.cpu.reg_sr, showFlags()).toUpperCase());
        textArea.append(String.format("A0:%08x  A1:%08x  A2:%08x  A3:%08x  USP:%08x\n",
                        this.cpu.addr_regs[0], this.cpu.addr_regs[1], this.cpu.addr_regs[2], this.cpu.addr_regs[3],this.cpu.reg_usp).toUpperCase());
        textArea.append(String.format("A4:%08x  A5:%08x  A6:%08x  A7:%08x  SSP:%08x\n\n",
                        this.cpu.addr_regs[4], this.cpu.addr_regs[5], this.cpu.addr_regs[6], this.cpu.addr_regs[7],this.cpu.reg_ssp).toUpperCase());
        this.sbuffer.setLength(0);
        int address = this.cpu.pc_reg*2;
        if(address < 0 || address >= this.cpu.readableSize())
        {
            this.sbuffer.append(String.format("%08x   ????", address));
        }
        else
        {
            CPU.DisassembledInstruction di =  getDI(address);
            di.shortFormat(sbuffer);
        }
        textArea.append(String.format("==> %s\n\n", sbuffer.toString().toUpperCase()));
    }

    /** 
     * Shows the CPU SR reg flag statusses.
     * 
     * @return a String with the flag status.
     */
    private String showFlags()
    {
        return  (this.cpu.isSupervisorMode()   ? "S " : " ")+ 
                    (((this.cpu.reg_sr&16)!=0) ? "X" : "-") + 
                    (((this.cpu.reg_sr&8)!=0)  ? "N" : "-") + 
                    (((this.cpu.reg_sr&4)!=0)  ? "Z" : "-") + 
                    (((this.cpu.reg_sr&2)!=0)  ? "V" : "-") + 
                    (((this.cpu.reg_sr&1)!=0)  ? "C" : "-");
    }
     
    /**
     * Gets a disassembled instruction.
     * @param address where to get the instruction from.
     * @return  the String with the disassembmed instruction.
     */
    private CPU.DisassembledInstruction getDI(int address)
    {
        int opcode = this.cpu.readMemoryWord(address);
        CPU.Instruction i = this.cpu.getInstructionFor(opcode);
        return i.disassemble(address, opcode,this.cpu);
    }            
    
     /**
     * Get the text area to display info/commands/results:
     * @param nbr the text area number either 1 (right) or 2 (left) - default=1.
     * @return the JTextArea chosen
     */
    private javax.swing.JTextArea getPanel(int nbr)
    {
        switch (nbr)
        {
            case 1:
            default:
                return this.dataLogger;
                
            case 2:
                return this.regLogger;
        }
    }
 

/*----------------------------Starting/Stopping of, and communication with, the emulation thread ---------------------------------*/
   
    
      /**
     * Starts the emulation.
     */
    public void startEmulation()
    {
        if (this.inifile.getTrueOrFalse("FAST-MODE"))
        {
            this.emulationThread= new smsqmulator.SimpleEmulationThread(this.cpu,this.gui.getScreen());
            this.emulationThread.start();
        }
        else
            goCommand(null,-1);
    }
    
    /**
     * This kills the emulation thread (and its daughter threads).
     */
    private void killEmulation()
    {
        if (this.emulationThread!=null && this.cpu !=null)
        {
            this.cpu.endEmulation();// kill the thread brutally whatever the consequences
            this.emulationThread=null;
        }
    }
    
    /**
     * Resets the emulator : restarts it entirely.
     */
    public void reset()
    {
        screens.Screen screen =  this.gui.getScreen();
        screen.fillScreen(0);
        screen.repaint();
        killEmulation();
        this.regLogger.setText("");
        this.dataLogger.setText("");
        this.lastCommand="";
        this.trapDispatcher.closeAllFiles();
        int memSize=inifile.getOptionAsInt("MEM_SIZE",1)*smsqmulator.Types.ONE_MiB;
        if (inifile.getTrueOrFalse("QL-SCREEN-EMULATION"))
            this.cpu = new CPU.CPUforScreenEmulation(memSize,screen,this.inifile,350000);// create the CPU, set its mem size & screen object
        else
            this.cpu = new CPU.MC68000Cpu(memSize,screen,this.inifile,350000);// create the CPU, set its mem size & screen object (350000 = max size of smsqe file)
        this.trapDispatcher.setCpu(this.cpu);
  
        boolean success = (SMSQE.loadRom(this.inifile.getOptionValue("ROM_IMAGE_FILE"), this.cpu) && SMSQE.CheckForValidSMSQE(this.cpu));// also sets PC and initial stack
        if (!success)
        {
            this.cpu=null;
            return;
        }
        success = SMSQE.setupSMSQE(this.cpu, this.inifile);
        if (success)
        {
            this.cpu.setTrapDispatcher(this.trapDispatcher);
            startEmulation();
        }
    }
    
    /**
     * This is called from the EmulationThread when it is done.
     * 
     * @param message a String to print out.
     */
    public synchronized void emulationThreadStopped(String message)
    {
        showInfo(false,this.diTextArea);
        this.regLogger.setText(message+"\n");
        this.emulationThread=null;
        if (this.inputWdw!=null)
        {
            this.inputWdw.requestFocus();
            this.inputWdw.requestFocusInWindow();
        }
    }
  
    /**
     * Suspends execution of the EmulationThread.
     * 
     * This is called from the Event Dispatch Thread, via the MonitorGui object when the window is minimized.
     * Note the use of <code>Thread.suspend</code> which is discouraged. However, adding a check for a "suspend yourself" flag within the emulationThread,
     * which is the recommended method, would slow the emulation down. Here there is no danger of the emulationThread and the
     * event dispatch thread blocking each other.
     */
    public void suspendExecution()
    {
        if (this.emulationThread!=null)
        {
            try
            {
               this.emulationThread.suspend();
            //    this.emulationThread.wait();
            }
            catch (Exception e)
            { /* NOP */ }
        }
    } 
    
    /**
     * Resumes execution of the EmulationThread.
     * 
     * This is called from the Event Dispatch Thread, via the MonitorGui object when the window is minimized.
     * Note the use of <code>Thread.resume</code> which is discouraged. However, adding a check for a "suspend yourself" flag within the emulationThread,
     * which is the recommended method, would slow the emulation down. Here there is no danger of the emulationThread and the
     * event dispatch thread blocking each other.
     */
    public void resumeExecution()
    {
        if (this.emulationThread!=null)
        {
            try
            {
                this.emulationThread.resume();
            }
            catch (Exception e)
            { /*NOP*/ }
        }
    }
    
    /**
     * Go into fast or slow mode
     * 
     */
    public void setFastMode()
    {
        killEmulation();
        try
        {
            Thread.sleep(500);
        }
        catch (Exception e){}
        startEmulation();
    }
    
    /**
     * Sets the interval between screen updates.
     * 
     * @param tim the interval between screen updates in milliseconds
     */
    public void setScreenUpdateInterval(int tim)
    {
        if (this.emulationThread!=null)
        {
            this.emulationThread.setScreenUpdateInterval(tim);
        }
    }
    
    /* ----------------- Go-between methods: communication between the Gui and others. These are called from the EDT -------------------*/
   
    /* ------------------------ Keuboard and mouse --------------------------*/
    /**
     * Set the Caps lock directly in the sysvars (!).
     * 
     * @param isSet  =<code> true</code> if caps lock status is to be set.
     */
    public void setCapsLockStatus(boolean isSet)
    {
        int t= isSet?0xff00:0;
        this.cpu.writeMemoryWord(0x28088, t);
    }
    
    /**
     * Input a key typed into the Java - SMSQE linkage block area.
     * <p> This should then be picked up by the SMSQE keyboard read routine.
     * 
     * @param key the "key" that was typed.
     */
    public void inputKey(int key)
    {
        this.trapDispatcher.addToKeyboardQueue(key);
    }
    
    /**
     * Inputs the mouse position into the SMSQE linkage area.
     * 
     *  *** This is called from the swing EDT via the MonitorGui object ***.
     * 
     * @param msmvtx  x movement of the mouse (how much did mouse move).
     * @param msmvty  y movement of the mouse.
     * @param msx   mouse x pos.
     * @param msy   mouse y pos.
     */
    public void inputMouse(int msmvtx,int msmvty,int msx,int msy)
    {
        if (this.cpu==null)
            return;
        msx=(msx<<16)+msy;                                  // mouse pos as long word
        msmvtx=(msmvtx<<16) | (msmvty&0xffff);              // mouse relatvie position
        this.cpu.writeMemoryLong(this.cpu.getLinkageBlock()+Types.LINKAGE_MOUSEPOS,msx);
        this.cpu.writeMemoryLong(this.cpu.getLinkageBlock()+Types.LINKAGE_MOUSEREL,msmvtx);
        this.trapDispatcher.resetCounter();
    }
    
    /**
     * Gets the mouse button pressed into the emulation.
     *  *** This is called from the swing EDT via the MonitorGui object ***.
     * 
     * @param btn  the button (1 = left, 2=right, 4 = middle button
     */
    public void inputMouseButton(int btn)
    {
        if (this.cpu==null)
            return;
        this.cpu.writeMemoryWord(this.cpu.getLinkageBlock()+Types.LINKAGE_MOUSEBTN,btn);
        this.trapDispatcher.resetCounter();
    }
    
    /**
     * Gets the mouse wheel movement into the emulation.
     *  *** This is called from the swing EDT via the MonitorGui object ***.
     * 
     * @param wheel  the wheel rotation and char.
     */
    public void inputMouseWheel(int wheel)
    {
        if (this.cpu==null)
            return;
        this.cpu.writeMemoryLong(this.cpu.getLinkageBlock()+Types.LINKAGE_MSEWHEEL,wheel);
        this.trapDispatcher.resetCounter();
    }
      
      /**
     * Sets a keyrow bit for the indicated row and col.
     * 
     * *** This is called from the swing EDT via the MonitorGui/Screen object ***
     * 
     * @param row row number.
     * @param col column number.
     */
    public void setKeyrow(int row,int col)
    {
        this.cpu.setKeyrow(row,col);
        this.trapDispatcher.resetCounter();                     // a key was pressed, down slow down immediatley
    }
 
    /**
     * Removes a keyrow bit for the indicated row and col.
     * 
     * *** This is called from the swing EDT via the MonitorGui/Screen object ***
     * 
     * @param row row number.
     * @param col column number.
     */
    public void removeKeyrow(int row,int col)
    {
        this.cpu.removeKeyrow(row,col);
        this.trapDispatcher.resetCounter();
    }   
    
    /**
     * Unsets all keyrows. 
     * This is used when the window is minimized (iconified). 
     * 
     */
    public void removeAllKeyrows()
    {
        this.cpu.removeAllKeyrows();
    }
    
    
    
    /**
     * Gets the CPU this object is monitoring.
     * 
     * @return the CPU this object is monitoring.
     */
    public CPU.MC68000Cpu getCPU()
    {
        return this.cpu;
    }
   
    /**
     * Gets the TrapDispatcher.
     * @return the TrapDispatcher.
     */
    public TrapDispatcher getTrapDispatcher()
    {
        return this.trapDispatcher;
    }
}

/*
//  M68k - Java Amiga MachineCore
//  Copyright (c) 2008-2010, Tony Headford
//  All rights reserved.
//
//  Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
//  following conditions are met:
//
//    o  Redistributions of source code must retain the above copyright notice, this list of conditions and the
//       following disclaimer.
//    o  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
//       following disclaimer in the documentation and/or other materials provided with the distribution.
//    o  Neither the name of the M68k Project nor the names of its contributors may be used to endorse or promote
//       products derived from this software without specific prior written permission.
//
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
//  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
//  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
//  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
//  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
*/
