
package smsqmulator;

/**
 * The independent thread that actually handles the emulation on a "g","gb" or "x" command.
 * 
 * The MonitorGoThread may observe breakpoints. When a breakpoint is hit, execution of the 68K program stops.<p>
 * The MonitorGoThread may log instructions. Logging is done to an internal data structure called a DebugList. Instructions executed are logged there
 * but are only written to the log file when execution of the program stops. The DebugList is a fifo queue and only keeps a certain
 * number of instructions (set up when this object is created).<p>
 * This thread may also watch a memory location and stop excution when the content of that memory location changes (long word only).<p>
 * This thread may also stop at a certain condition (eg register==value).
 * <p>
 * In "<code>gofast</code>" mode, no memory is watched, no breakpoints are checked etc - the program just executes as fast as it can.
 * 
 * @author and copyright (c) 2012 - 2019 Wolfgang Lenerz
 * @version 
 *  1.03 Simplified constructor parameters.
 *  1.02 DebugList uses array.
 *  1.01 set TrapDispatcher for CPU even if slow mode.
 */
 public class MonitoredEmulationThread  extends Thread implements EmulationThread
 {
    private final CPU.MC68000Cpu cpu;                           // CPU to be used in emulation thread
    private final Monitor monitor;                              // monitor for this thread
    private QL50HzInterrupt ih =null;                           // external interrupt handler
    private final TrapDispatcher trapDispatcher;                // TrapDispatcher used to dispatch traps to drivers & callback traps to java.
    private int iterations;                                     // how many instructions are to be executed?
    private int deduct=1;                                       // number deducted from the number of instructions to be executed.
    private boolean stopNow;                                    // as soon as this is <code>true</code>, the thread terminates (e.g. at a breakpoint).
    private DebugList dbl=null;                                 // inner class, the list containing the debug info for logging
    private final StringBuilder sBuilder;                       // used for building logging info
    
    private boolean watchBreakpoints=false;                     // true if breakpoints need to be watched
    private final java.util.ArrayList<Integer> breakpoints;     // the array with the breakpoints.
    private int provbkp = -1;                                   // provisional breakpoint used with traceover
   
    private int conditionDataReg=-1;                            // data reg to check for conditional break
    private int conditionAddressReg=-1;                         // address reg to check for conditional break
    private int conditionValue;                                 // the value, for a conditonal break.
    private boolean conditionIsContent;                         // only check condition if this is true
    private int conditionAddress=-1;                            // address to check for condition
    private boolean signalDeath=true;                           // whether the death of the thread should be shown in the MonitorGui log wdw
  
    private boolean ckmem=false;                                // true if memory is to be watched (i.e. change of mem address content)
    private volatile int oldval=0;                              // old value of memory to be watched
    private int memWatched=0;                                   // what mem address is to be watched
 
    private boolean logInstructions=false;                      // true if instructions should be logged
    private int nbrLogs;                                        // how many instructions we log
    private boolean excludeSuper=false;                         // when logging, exclude all instructions executed in supervisor mode
    private String instructionsListFilename="";                 // the file into which the logged instructions will be dumped
  //  private final boolean oldSupervisor=false;                        // used when saving logged instructions, signals SV mode change
    private int upperLogLimit=0;                                // when logging instructions, instructions beyond this limit will NOT be logged, set to 0 to log everything (useful to exclude OS code in high mem).

    /**
     * Creates the monitor go thread.
     * 
     * @param nbrOfInstructions how many instructions are to be executed, or 0 if continuous execution.
     * @param monitor the motor launching this goThread.
     * @param watchBreakpoints <code>true</code> if breakpoints should be observed.
     * @param upperLogLimit when logging instructions, instructions beyond this limit will NOT be logged (useful to exclude OS code in high mem).
     * @param checkmem set to <code>true</code> if memory is to be watched.
     * @param memWatched what memory address is to be watched.
     * @param provbreak provisional breakpoint used with traceover.
     * @param excludeSuper if <code>true</code> : when logging, exclude all instructions executed in supervisor mode.
     * @param screen the screen currently used to display the emulation.
     * @param loginstrs is true if instructions need to be logged.
     * @param nbrLogs how many instructions should be logged
     */
    public MonitoredEmulationThread (int nbrOfInstructions,Monitor monitor, boolean watchBreakpoints,int upperLogLimit,boolean checkmem,
                                     int memWatched,int provbreak, boolean excludeSuper,screens.Screen screen, boolean loginstrs, int nbrLogs)
                           
    {
        if (nbrOfInstructions!=0)
        {
            this.iterations=nbrOfInstructions;
            this.deduct=1;
        }
        else
        {
            this.iterations=1;
            this.deduct=0;
        }
        this.monitor=monitor;
        this.cpu=monitor.getCPU();
        this.breakpoints=monitor.getBreakpoints();
        String filename=monitor.getLogfilename();
        if (filename!=null && !filename.isEmpty())
            this.instructionsListFilename=filename;             // name of file to (possibly) save instructions to.
        this.nbrLogs = nbrLogs;
        this.logInstructions=loginstrs;
        this.watchBreakpoints=watchBreakpoints;
        this.trapDispatcher= monitor.getTrapDispatcher();
        this.sBuilder=new StringBuilder(400);
        this.upperLogLimit=upperLogLimit;
        this.ckmem=checkmem;
        this.memWatched=memWatched;
        if (memWatched>0)
            this.oldval=this.cpu.readMemoryLong(memWatched);
        this.provbkp=provbreak;
        this.excludeSuper=excludeSuper;
        this.dbl=new DebugList(this.nbrLogs);
        this.setName("Main emulation thread (monitored)");
        
        this.ih = new QL50HzInterrupt(this.cpu);
        this.ih.setScreen(screen);
    }
   
    /**
     * Does the actual work : execute cpu instructions in a loop.
     */
    @Override
    public void run()
    {
        this.dbl.clear();
        if (this.ih!=null)
        { 
            this.ih.startInterrupts();                      // this stops interrupt threads before starting new ones
        }
        done(execSomeInstructions());
    }
    
    /**
     * Exec one or several instructions.
     * 
     * @return error message or ""
     */
    private String execSomeInstructions()
    {
        boolean addressFlag=false;
        int execResult;  
        String message="";
        while (this.iterations>0 && !this.stopNow)
        {
            try
            {
                if (this.logInstructions)                   // are instructions to be logged ?
                {
                    if (!(this.excludeSuper && cpu.isSupervisorMode()))
                    {
                        if (this.upperLogLimit!=0)
                        {
                            int addr = this.cpu.pc_reg*2;       // if there is an upper limit, don't log if code we're executing is above that
                            if (addr<this.upperLogLimit)
                                saveInstruction();
                        }
                        else
                            saveInstruction();
                    }
                }
                if (this.cpu.pc_reg*2==this.conditionAddress)
                    addressFlag=true;
                execResult = this.cpu.execute();
                if (execResult<0)
                {
                    switch (execResult)
                    {
                        case -10:                               // "Hard" breakpoint
                            this.stopNow=true;
                            message="Hard breakpoint";
                            return message;
                        case -15:
                            return "";                          // kill emulation
                  
                        case -20:                               // hit an illegal istruction!
                            this.stopNow=true;
                            message="ILLEGAL instruction!";
                            return message;
                    
                        case -30:                               // used whenever I want to save logged instructions to file
                            saveInstruction();
                            this.dbl.saveToFile(this.instructionsListFilename);
                            break;
                            
                        default:
                            this.trapDispatcher.dispatchTrap(-execResult,this.cpu);// it was some kind of TRAP instruction, or another request to come back to Java
                            break;
                    }
                }
                int addr = this.cpu.pc_reg*2;
                this.cpu.stopNow=0;
                if (this.upperLogLimit!=0)                         // only count iterations if PC is below some arbitrary limit
                {                                               // (used to mean that when G is done, interrupts & trap are NOT counted)
                    if (addr<this.upperLogLimit)
                        this.iterations-=this.deduct;
                }
                else
                    this.iterations-=this.deduct;
                if (this.ckmem)                                 // are we watching for a memory change?
                {
                    if(this.cpu.readMemoryLong(this.memWatched)!=this.oldval)
                    {
                        this.stopNow=true;
                        message="Memory changed! - old value = "+this.oldval;
                        this.oldval=this.cpu.readMemoryLong(this.memWatched);
                        message+=" new value = "+this.oldval;
                    }
                }
                if (this.provbkp!=-1)                           // check if there is a provisional breakpoint to be honored
                {
                    if (addr==this.provbkp)
                    {
                        this.stopNow=true;
                        message="Provisional Breakpoint hhit";
                        break;
                    }
                }
                if (this.watchBreakpoints && this.breakpoints!=null)// any other breakpoint to be honored?
                {
                    if(breakpoints.contains(addr))              // is this on a breakpoint?
                    {
                        this.stopNow=true;
                        message="Permanent breakpoint hit";
                        break;
                    }
                }
                if (checkConditon(addressFlag))                 // check for condition
                {
                    this.stopNow=true;
                    message="Condition met";
                }
            }
            catch(Exception e)
            {
                this.stopNow=true;
            }
        }
        return message;
    }
    
    /**
     * This is called when the emulation stops.
     * 
     * @param message the message to display.
     */
    public void done(String message)
    {
        if (this.ih!=null)
            this.ih.stopInterrupts();                           // stop the interrupt handler
            
        this.provbkp=-1;                                        // no more provisional breakpoint
        if (this.signalDeath)
        {
            this.monitor.showInfo(true,null);                   // show death of this thread
            this.monitor.emulationThreadStopped(message);
        }
        if (this.logInstructions)
            this.dbl.saveToFile(this.instructionsListFilename);
    }

     /**
     *
     */
    @Override
    public void endEmulation()
    {
        if (this.ih!=null)
        {
            this.ih.stopInterrupts();
            this.ih.setScreen(null);
            this.ih=null;
        }
        this.stop();
    }
    
    /** 
     * Switches logging state on/off.
     * 
     * @param newLogState if true, logging is switched on again
     */
    @Override
    public synchronized void switchLogging(boolean newLogState)
    {
        this.logInstructions=newLogState;
        if (!newLogState)                                       // logging was switched off, save instructions
        {
            this.dbl.saveToFile(this.instructionsListFilename);
        }
        else
        {
            if (this.nbrLogs!=0)
                this.dbl.clear();                                   // logging switched on, clear list
            else
            {
                this.nbrLogs=50000;
                this.dbl=new DebugList(this.nbrLogs);
            }
        }
    }
    
    /**
     * Do we log calls to the OS or only user state code?
     * 
     * @param superState if true, only log calls in user mode.
     */
    public synchronized void switchSuper(boolean superState)
    {
        this.excludeSuper=superState;
    }
    
    /**
     * "Saves" an instruction into the DebugList.
     */
    private void saveInstruction()
    {
        this.sBuilder.setLength(0);
        ListElement le=new ListElement();                
        le.instrAddr = this.cpu.pc_reg*2;
        le.instrAddrContent = this.cpu.readMemoryLong(le.instrAddr);
        for (int i=0;i<8;i++)
        {
            le.regsData[i]=this.cpu.data_regs[i];
            le.regsAddr[i]=this.cpu.addr_regs[i];               // set data and addr regs
        }
        int A7 = this.cpu.addr_regs[7];
        for (int i=0; i<le.stackContent.length;i++,A7+=4)       // content on stack
        {
            le.stackContent[i]= cpu.readMemoryLong(A7);
        }
        
        if(le.instrAddr > 0)
        {
            int opcode =  this.cpu.readMemoryWord(le.instrAddr);     // opcode at current address
            CPU.Instruction i = this.cpu.getInstructionFor(opcode);
            le.di = i.disassemble(le.instrAddr, opcode,this.cpu);
        }
        le.regStatus=this.cpu.reg_sr;
        this.dbl.add(le);
    }
    /**
     * "Saves" an instruction into the DebugList.
     *
    private void saveInstructionOLD()
    {
        this.sBuilder.setLength(0);
        ListElement le=new ListElement();                       // uses inner class 
        int address = cpu.pc_reg*2;
        le.memaddress=(String.format("%08x ",address));

        if(address < 0)
        {
            sBuilder.append(String.format("%08x   ????", address));
        }
        else
        {
            int opcode =  this.cpu.readMemoryWord(address);     // opcode at current address
            CPU.Instruction i = this.cpu.getInstructionFor(opcode);
            CPU.DisassembledInstruction di = i.disassemble(address, opcode,this.cpu);
            boolean showBytes=false;
          
            this.sBuilder.append(String.format("  %08x ", cpu.readMemoryLong(address)));
        }
        le.instruction=this.sBuilder.toString();
        
        int a7=this.cpu.addr_regs[7] ;                         
        this.sBuilder.setLength(0);
        this.sBuilder.append(String.format("A7: %08x  ",a7));
       // a7&= 0xfffffffe;                                        // just in case...
        for (int i=0;i<4;i++)
        {
            this.sBuilder.append(String.format("    %08x  ",cpu.readMemoryLong(a7+(i*4))));
        }
        le.memcontent=this.sBuilder.toString();
        
        this.sBuilder.setLength(0);
        for (int i=0;i<8;i++)
        {
            this.sBuilder.append(String.format("D"+i+": %08x  ",this.cpu.data_regs[i]));
        }
        le.regContentsData=this.sBuilder.toString();
        
        this.sBuilder.setLength(0);
        for (int i=0;i<8;i++)
        {
            this.sBuilder.append(String.format("A"+i+": %08x  ",this.cpu.addr_regs[i]));
        }
        le.regContentsAddr=this.sBuilder.toString();
        
        if (this.cpu.isSupervisorMode()!=this.oldSupervisor)
        {
            this.oldSupervisor=!this.oldSupervisor;
            le.supervisor="New Supervisor state = "+this.oldSupervisor;
        }
        else
        {
            le.supervisor="";
        }
        this.dbl.add(le);
    }
    */
    
    /**
     * Checks for a condition at a certain address.
     * 
     * @param addressFlag
     * 
     * @return 
     */
    private boolean checkConditon(boolean addressFlag)
    {
        if (this.conditionIsContent)                            // the condition is for the content of an address pointed to by add reg
        {
            if (this.conditionAddressReg==-1)                   // no addr reg?
                return false;
            if (!addressFlag)                                   // we're not at the right address
                return false;
            return this.conditionValue == this.cpu.readMemoryLong(this.cpu.addr_regs[this.conditionAddressReg]);// get value & compare
        }
        
        if (this.conditionAddressReg!=-1)
        {
            return this.cpu.addr_regs[this.conditionAddressReg]==this.conditionValue;
        }
        if (this.conditionDataReg!=-1)  
        {
           return this.cpu.data_regs[this.conditionDataReg]==this.conditionValue;
        }
        return false;                                           // what is this condition?
    }
    
    /**
     * Sets the condition for a "g" command
     * @param dataReg the datareg this condition applies to, or -1 if condition is for address reg.
     * @param addrReg the addrreg this condition applies to, or -1 if condition is for data reg.
     * @param condValue the value to be met.
     * @param isContent possible content at...
     * @param address ...this address
     */
    @Override
    public synchronized void setCondition (int dataReg,int addrReg,int condValue,boolean isContent,int address)
    {
        this.conditionValue=condValue;
        this.conditionDataReg=dataReg;
        this.conditionAddressReg=addrReg;
        this.conditionIsContent=isContent;
        this.conditionAddress=address;
    }
    
    /**
     * Sets whether the death of the thread should be logged in the MonitorGui log wdw.
     * 
     * @param b <code>true</code> if death should be logged.
     */
    public synchronized void signalDeath(boolean b)
    {
        this.signalDeath=b;
    }
    
    /**
     * Sets whether memory check should be switched on or off.
     * 
     * @param b true if memory check is the switched on.
     */
    @Override
    public synchronized void setckmem(boolean b)
    {
        this.ckmem=b;
    }
    
    /**
     * Sets the memory to be watched.
     * 
     * @param a the mmemory to be watched : if content of memory location is changed the emulation breaks (in monitor mode).
     */
    @Override
    public synchronized void setmemWatched(int a)
    {
        this.oldval=this.cpu.readMemoryLong(a);
        this.memWatched=a;
    }
     
    /**
     * Sets the upper limit of memory when logging.
     * 
     * @param address the upper limit of memory when logging any instruction executed above this will not be logged.
     */
    @Override
    public synchronized void setUpperLimit(int address)
    {
        this.upperLogLimit=address;
    }
    
    
    /**
     * Sets the interval between screen updates.
     * 
     * @param tim the interval between screen updates in milliseconds.
     */
    @Override
    public void setScreenUpdateInterval(int tim)
    {
        if (this.ih!=null)
            this.ih.setScreenUpdateInterval(tim);
    }
    /*---------------------------- classes for debugging --------------------------------------*/

  
   
    /**
     * Element of debuglist, just data
     */
    private class ListElement
    {
        public CPU.DisassembledInstruction di;
        public int instrAddr;
        public int instrAddrContent;
        public int[] stackContent=new int[5];
        public int[] regsData=new int[8];
        public int[] regsAddr=new int[8];
        public int regStatus;
        public String supervisor;
    }
    
    /**
     * A round robin list : elements are added at end, if list size reached, the first element is discarded
     */
    private class DebugList
    {
        private int last;
        private final int maxsize;
        private final ListElement[] elmts;
        private static final String REGS_FILLER="                                            ";
        private int previousSR;
        
        public DebugList(int maxsize)
        {
            this.maxsize=maxsize;
            this.elmts=new ListElement[maxsize];
        }
        
        public void clear()
        {
            this.last=-1;
        }
        
        public void add(ListElement el)
        {
            if (this.last!=-1 && el.instrAddr == (this.elmts[this.last].instrAddr))
                return;                                         // sometimes an interrupt may cause an instruction to be treated twice  
            this.last++;
            if (this.last>=this.maxsize)
                this.last=0;   
            this.elmts[this.last]=el;
        }
        /**
         * Save list to file
         * @param filepath 
         */
        public void saveToFile(String filepath)
        {
            ListElement el;
            int index=0;
            try 
                (
                    java.io.PrintWriter out = new java.io.PrintWriter(new java.io.FileWriter(filepath)); 
                    java.io.PrintWriter out2 = new java.io.PrintWriter(new java.io.FileWriter(filepath+"2"));
                )
            {
                StringBuilder sb = new StringBuilder(200);
                int ptr=this.last+1;                            // last +1 = first!
                if (ptr>=this.maxsize)
                    ptr=0;
                while (ptr!=this.last)
                {
                    sb.setLength(0);
                    this.elmts[ptr].regStatus &= CPU.MC68000Cpu.SUPERVISOR_FLAG;
                    index=saveElementAsFormattedString(out,out2,this.elmts[ptr],index,this.previousSR,sb);
                    this.previousSR = this.elmts[ptr].regStatus;
                    ptr++;
                    if (ptr>=this.maxsize)
                        ptr=0;
                }
                saveElementAsFormattedString(out,out2,this.elmts[this.last],index,this.previousSR,sb);
                out.close();
                out2.close();
            }
            catch (Exception e)
            {
            }
        }
        
        /**
         * Prints an element.
         * 
         * @param out
         * @param out2
         * @param el
         * @param index
         * 
         * @return a code indicating how many spaces should be prepended before line, to make indentation.
         */
        private int saveElementAsFormattedString(java.io.PrintWriter out,java.io.PrintWriter out2,ListElement le,int index, int previousSVState,StringBuilder sb)
        {
            if (le==null)
                return index;
            String pad="                                                                                                                                         "+
                    "  ";  
            
            String temp;
            if (index>0 && index<pad.length()-1)
                temp=pad.substring(0,index);                    // make indentation
            else
                temp="";
                                                                // contents of regs before the instructions
            makeArrayString(le.regsData,"D",sb);
            out.println(temp+DebugList.REGS_FILLER+sb.toString()); // data regs
            
            sb.setLength(0);
            makeArrayString(le.regsAddr,"A",sb);
            out.println(temp+DebugList.REGS_FILLER+sb.toString()); // address regs
 
            sb.setLength(0);
            sb.append(temp).append(DebugList.REGS_FILLER).append("(A7): ");// content of A7 stack
            makeArrayString(le.stackContent,"",sb);
            out.println(sb.toString());
            
            sb.setLength(0);
    //        sb.append(String.format("%08x ",le.instrAddr));                                  // instruction
            if (le.di!=null)
                le.di.shortFormat(sb);                          //* if(showBytes)   di.formatInstruction(this.sBuilder);
            else
                sb.append(String.format("%08x   ????", le.instrAddrContent));
            
            sb.append(String.format("   %08x   ",le.instrAddrContent));
            String instr = sb.toString();
            out.println(instr);
            out2.println(instr);
            
            if (previousSVState != le.regStatus)
            {
                String status = le.regStatus!=0?"SV mode SET":"No SV mode";
                out.println("  SUPERVISOR CHANGE: New status " + status);
                out2.println("  SUPERVISOR CHANGE: New status " + status);
            }
            if ((instr.contains("bsr"))|| (instr.contains("jsr")))
                index+=2;
            if ((instr.contains("rts"))|| (instr.contains("rte")))
                index-=2;
            if (index<0)
                index=0;
            return index;
        }
        
        /**
         * Add the formatted contents of the array to the StringBuilder
         * 
         * @param array     array with data
         * @param prepend   possible header for each array element
         * @param sb        StringBuilder to which the formatted content will be appended.
         */
        private void makeArrayString(int[] array,String prepend,StringBuilder sb)
        {
            for (int i = 0;i<array.length;i++)
            {
                if (prepend!=null && !prepend.isEmpty())
                    sb.append(prepend).append(i).append(": ");
                sb.append(String.format("%08x  ",array[i]));
            }
        }
    }


}
