/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package smsqmulator;

/**
 * Methods to setup smsqe.
 * 
 * @author and copyright (c) Wolfgang Lenerz 2023.
 * 
 * @version 1.00
 */
public class SMSQE
{
    /**
     * Checks that SMSQE is correct, sets many of the different config options.
     * 
     * @param cpu       what cpu loaded the rom file
     * @param iniFile   where to find config options
     * 
     * @return  true if SMSQE setup ok.
     * 
     * NB split off from CPU.
     */
    public static boolean setupSMSQE(CPU.MC68000Cpu cpu, inifile.IniFile iniFile)
    {
        int version=cpu.readMemoryLong(cpu.getTotalRamSize()+0x30);                     // major version, corresponds to SMQE release         
        if (version<smsqmulator.Helper.convertStringToInt(smsqmulator.Types.MINIMUM_VERSION_NEEDED))     
        {                                                   // not found or too old
            smsqmulator.Helper.reportError(ui.Localization.Texts[45], ui.Localization.Texts[103], null);
            return false;
        }
        int ramSize = cpu.getRamSizeInMiB()*(1024*1024);
        int totRamSize = cpu.getTotalRamSize();
        cpu.writeMemoryLong(0x1000a,0x476f6c64);                // ******************************************** necesaary for some c progs (why?)
        cpu.writeMemoryLong(ramSize-4,ramSize);                 // show top of ram
        cpu.writeMemoryLong(4,ramSize);                         // address at $4 points to linkage block
        
        int temp;
        String s=iniFile.getOptionValue("NFA_USE");             // config NFA etc use name
        if (!s.isEmpty())
        {
            temp=smsqmulator.Helper.convertUsageName(s);
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_NFA_USE,temp);
        }
        s=iniFile.getOptionValue("SFA_USE");     
        if (!s.isEmpty())
        {
            temp=smsqmulator.Helper.convertUsageName(s);
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_SFA_USE,temp);
        }
        s=iniFile.getOptionValue("WIN_USE");
        if (!s.isEmpty())
        {
            temp=smsqmulator.Helper.convertUsageName(s);
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_WIN_USE,temp);
        }
        s=iniFile.getOptionValue("FLP_USE");
        if (!s.isEmpty())
        {
            temp=smsqmulator.Helper.convertUsageName(s);
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_FLP_USE,temp);
        }
 /*       s=iniFile.getOptionValue("MEM_USE");
        if (!s.isEmpty())
        {
            temp=smsqmulator.Helper.convertUsageName(s);
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_MEM_USE,temp);
        }
 */       
        cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_BOOT_DEVICE,smsqmulator.Types.WINDriver+1);//boot device
        cpu.writeMemoryWord(ramSize+smsqmulator.Types.LINKAGE_RANDOM,new java.util.Random().nextInt(65535));// set next random number
      
        // now find out whether the ROM loaded is OK.
        int addr=cpu.findInMemory(totRamSize,totRamSize+1000,"SMSQXqXq");// find my marker
        if (addr==-1)
        {                                                   // not found : error, rom is too old
            smsqmulator.Helper.reportError(ui.Localization.Texts[45], ui.Localization.Texts[103], null);
            return false;
        }
 
        if (version == smsqmulator.Helper.convertStringToInt(smsqmulator.Types.MINIMUM_VERSION_NEEDED))// check minor version if both majors are the same
        {
            version=cpu.readMemoryLong(addr+8);                        // minor version, corresponds to SMSQmulator specific SMSQE release (if any)  
            if (version<smsqmulator.Helper.convertStringToInt(smsqmulator.Types.MINIMUM_MINOR_VERSION_NEEDED))     
            {                                                   // not found or too old
                smsqmulator.Helper.reportError(ui.Localization.Texts[45], ui.Localization.Texts[103], null);
                return false;
            }
        }
        
        // set "slower  when idle flag"
        int slower  = iniFile.getOptionAsInt("LESS-CPU-WHEN-IDLE",1);
        cpu.writeMemoryByte(addr + 12 + 0x16,slower);
        
        
        // the next parts find and set whether devices should be linked into SMSQE
        // !!!!!!!!!!!!!!!!!!!!!! KEEP THEM IN THE CORRECT ORDER §§§§§§§§§§§§§§§§§§§§§§
        addr=setOption(addr, totRamSize, "WIN0", "DISABLE-WIN-DEVICE", cpu, iniFile);        
        addr=setOption(addr, totRamSize, "MEM0", "DISABLE-MEM-DEVICE", cpu, iniFile);        
        addr=setOption(addr, totRamSize, "NFA0", "DISABLE-NFA-DEVICE", cpu, iniFile);        
        addr=setOption(addr, totRamSize, "SFA0", "DISABLE-NFA-DEVICE", cpu, iniFile);        
        setOption(addr, totRamSize, "FLP0", "DISABLE-FLP-DEVICE", cpu, iniFile);
        return true;
    }
    
    /**
     * Load the rom file into the cpu memory.
     * 
     * @param filename  name of entire path to file.
     * @param cpu       the cpu into which the rom is to be loaded.
     * 
     * @return true if rom was loaded ok
     */
    public static boolean loadRom(String filename, CPU.MC68000Cpu cpu)
    {
        if (filename==null)
            return false;
        try (java.nio.channels.FileChannel inChannel = new java.io.RandomAccessFile(filename, "r").getChannel() )
        {   
            int fileSize=(int)inChannel.size();
            cpu.readFromFile(cpu.getTotalRamSize(),fileSize,inChannel,true);// get rom image into memory 
            return true;
        }
        
        catch (java.lang.OutOfMemoryError e)
        {
            smsqmulator.Helper.reportError(ui.Localization.Texts[45], ui.Localization.Texts[108], null); // inform user : not enough memory
            return false;                                   // OOOOPs out of memory
        }
        catch (Exception e2)
        {
            smsqmulator.Helper.reportError(ui.Localization.Texts[45], ui.Localization.Texts[50], null, e2);//inform user : can't read smsqe file
            return false;
        }
    }
   
    /**
     * Checks that SMSQ/E is valid (the right version) and prepares the CPU for starting execution (sets the initial stack pointer and pc)
     * 
     * @param cpu the cpu to check
     * 
     * @return  true if valid SMSQE
     */
    public static boolean CheckForValidSMSQE(CPU.MC68000Cpu cpu)
    {
        int temp, addr;
        int ramSize = cpu.getRamSize();
        int totRamSize = cpu.getTotalRamSize();
        int totMemSize = cpu.getTotalMemSize();
        
        char[] mainMemory = cpu.getMemory();
        if (cpu.screen!=null)
        {
            cpu.screen.fillScreen(0);
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_SCREENBASE,cpu.getScreenStart());
            cpu.writeMemoryLong(ramSize+smsqmulator.Types.LINKAGE_SCREENSIZE,cpu.screen.getScreenSizeInBytes());
            mainMemory[(ramSize+smsqmulator.Types.LINKAGE_SCREEN_LINE_SIZE)/2]=(char)cpu.screen.getLineSize();
            mainMemory[(ramSize+smsqmulator.Types.LINKAGE_SCREEN_XSIZE)/2]=(char)cpu.screen.getXSize();
            mainMemory[(ramSize+smsqmulator.Types.LINKAGE_SCREEN_YSIZE)/2]=(char)cpu.screen.getYSize();       //xxxxxxxxxxxxx
           
            // find the config info & set screen mode in it
            addr=cpu.findInMemory(totRamSize,totMemSize,smsqmulator.Types.SMSQMULATOR_CONFIG_FLAG,"SMSQ");// search in "ROM"
            if (addr==-1)
            {
                smsqmulator.Helper.reportError(ui.Localization.Texts[45], ui.Localization.Texts[103], null);// oops, this smsqe is too old
                return false;
            }
            else
            {
                int tx=addr+smsqmulator.Types.SMSQMULATOR_CONFIG_FLAG.length*2+4+12+10; // jump over general config flag, special config flag (4) ,version info  (12) and into the config block (10)
                tx/=2;
                temp=  (mainMemory[tx]&0xff00) +cpu.screen.getColourMode();
                mainMemory[tx]=(char)(temp&0xffff);   // set screen mode in config block  //xxxxxxxxxxx
            }
            cpu.pc_reg=cpu.getTotalRamSize()/2;                   // set PC to start of "ROM" = image that was loaded
            cpu.reg_ssp=cpu.ramSize-8;                            // set SSP to top of usable ram
            cpu.addr_regs[7]=cpu.reg_ssp;      
            cpu.reg_sr = 0x2700;  
            return true;
        }
       return false;                         // if there is no screen something funny happened
    }
    
    private static int setOption(int addr, int totRamSize,String marker, String configName,CPU.MC68000Cpu cpu,inifile.IniFile iniFile )
    {
        if (addr==-1)
            return -1;
        addr=cpu.findInMemory(addr,totRamSize,smsqmulator.Types.SMSQMULATOR_CONFIG_FLAG,marker);
        if (addr!=-1)
        {
            addr+=marker.length() + smsqmulator.Types.SMSQMULATOR_CONFIG_FLAG.length*2;
            if (iniFile.getTrueOrFalse("DISABLE-WIN-DEVICE"))
                cpu.writeMemoryByte(addr,1);
            else
                cpu.writeMemoryByte(addr,0);    
        }  
        return addr;
    }
}
