/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package smsqmulator;

import CPU.MC68000Cpu;

/**
 * Methods to get string from smsqe to java and vice-versa.
 * 
 * @author and copyright (c) 2012-2023 Wolfgang Lenerz
 * @version 1.00
 */
public class SmsqeString
{
    
    private final static String JAVA_ACCENTED ="@[]£{|}~äéöüçœàñáâëèêïíîóôùúûßÄÉÖÜÇøå¿æÑÆαŒ€"; // these java accented chars will be converted to ...
    private final static int [] SMSQE_ACCENTED={64,91,93,96,123,124,125,126,128,131,// ...these SMSQE chars and vice-versa.
                       132,135,136,139,141,137,140,142,143,144,145,146,147,149,
                       150,152,154,153,155,156,160,163,164,167,168,134,130,180,
                       138,169,170,172,171,181};
    
   /**
    * Returns a java String from an SMSQE string at a certain address MAKING accented chars conversion.
    * 
    * @param address where the string lies in memory. This is presumed to be even (after all, the string is preceded by a length word).
    * @param cpu the cpu with the momory
    * 
    * @return the String read, this may be "" but will not be <code>null</code>.
    * <p> This makes character conversion from SMSQQE charset to Java charset.
    */
    public static final String readSmsqeString(int address,CPU.MC68000Cpu cpu)
    {
        address&=CPU.MC68000Cpu.CUT_OFF;                        // !!! remove higher bits  also means that address won't be <0
        int xxx = address;
        char[]mem = cpu.getMemory();
        address/=2;                                             // word sized mem access
        if (address>=mem.length)
            return"";
        int count=mem[address];                               // nbr of bytes in string
        address++;                                              // jump over length word
        if (count==0 || ((address+count)>=mem.length))// don't go above max ROM address
            return "";                             
        StringBuilder result=new StringBuilder(count);
        int res;
        for (int i=0;i<count;i+=2,address++)
        {   
            res=mem[address];
            result.append(convertToJava((byte)(res>>>8)));
            result.append(convertToJava((byte)(res&0xff)));
        }
        if ((count & 1) == 1)                               // string was one too long
            result.setLength(result.length()-1);
        return result.toString();
    }
    
    /**
    * Writes a java String as an SMSQE string to a certain address MAKING accented chars conversion.
    * 
    * @param address - where to start writing.
    * @param s the <code>String</code> to write.
    * @param maxLength the max length of the string to write NOT INCLUDING THE LENGTH WORD, -1 if no max length.
    * @param cpu the cpu with the momory
    */
    public static final void writeSmsqeString(int address,String s,int maxLength,CPU.MC68000Cpu cpu)
    {
        writeSmsqeString(address,s,true,maxLength,cpu);
    }
   
   /**
    * Writes a java String as an SMSQE string to a certain address MAKING accented chars conversion.
    * 
    * @param address - where to start writing.
    * @param s the <code>String</code> to write.
    * @param writeLength <code>true</code> if the length of the string should be prepended to the string (as would be usual in SMSQE)
    * @param maxLength the max length of the string to write NOT INCLUDING THE LENGTH WORD, -1 if no max length.
    * @param cpu the cpu with the momory
    */
    public static final void writeSmsqeString(int address,String s,boolean writeLength,int maxLength,CPU.MC68000Cpu cpu)
    {   
        if(s==null)
            return;                                         // strange string, or writing normal string to screen mem : that doesn't make sense here!
        address&=MC68000Cpu.CUT_OFF;
        char[] mem=cpu.getMemory();
        address/=2;                                         // memory is array of shor
        if  (s.isEmpty())
        {
            if (writeLength)
                mem[address]=0;
            return;
        }
        int count=s.length(); 
        if ((maxLength!=-1) && (count>maxLength))
        {
            count=maxLength;
        }
        if (((1+address+count/2)>=mem.length))            // don't write above max ROM address
            return;
        boolean odd=(count&1)==1;
        if (writeLength)
            mem[address++]=(char)count;                     // xxxxxxxxxx
        count&=0xfffffffe;                                  // make even if odd count
        char res,res1;
        for (int i=0;i<count;i+=2,address++)
        {
            res=(char)convertToSMSQE(s.charAt(i)); // convert char if need be
            res<<=8;
            res1=(char) (convertToSMSQE(s.charAt(i+1)));
            res|=(res1&0xff);
            mem[address]=res;
        }
        if (odd)
        {
            count=(byte)convertToSMSQE(s.charAt(s.length()-1));
            count<<=8;
            res=mem[address];
            res&=0xff;
            mem[address]=(char)(count|res);
        }
    }
    
     /**
     * Converts from java string to smsqe char/byte.
     * 
     * @param c the char to convert
     * 
     * @return the converted char
     */
    public static final byte convertToSMSQE(String c)
    {
        // assert Helper.javaAcc.length()==SMSQE_ACCENTED.length;
        if (c==null || c.isEmpty())
        {
            return 0;
        }
        if (c.length()!=1)
            c=c.substring(0,1);
        int p= JAVA_ACCENTED.indexOf(c);
        if (p!=-1)
            return (byte)(SMSQE_ACCENTED[p]&0xff);
        else 
            return (byte) c.charAt(0);
    }
    
    /**
     * Converts from java char to smsqe char/byte.
     * 
     * @param c the char to convert
     * 
     * @return the converted char
     */
    public static final byte convertToSMSQE(char c)
    {
        assert JAVA_ACCENTED.length()==SMSQE_ACCENTED.length;
        int p= JAVA_ACCENTED.indexOf(c);
        if (p!=-1)
            return (byte)(SMSQE_ACCENTED[p]&0xff);
        else 
            return (byte) c;
    }
    
    /**
     * Converts from SMSQE byte to Java string.
     * 
     * @param c the byte ti convert
     * 
     * @return the byte as a java unicode string.
     */
    public static final String convertToJava(byte c)
    {
        assert JAVA_ACCENTED.length()==SMSQE_ACCENTED.length;
        int in=(int)c&0xff;
        for (int i=0;i<SMSQE_ACCENTED.length;i++)
        {
            if (SMSQE_ACCENTED[i]==in)
                return JAVA_ACCENTED.substring(i,i+1);
        }
        return Character.toString((char)c);
    }
    
    
    //  DO NOT DELETE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    //  This it may be used when debugging a WinFile
    /** 
    * Gets a java String from an SMSQE string in a ByteBuffer MAKING accented chars conversion.
    * 
    * @param buffer the buffer to get the String from.the <code>String</code> to write.
    * @param offset - where to start reading in the buffer
    * 
    * @return the <code>String</code> read. May be <code>null</code>.
    * 
    * NB used for debugging, see winfile.
    *  
  
    public static final String readSmsqeString2(java.nio.ByteBuffer buffer,int offset)
    {
        byte l=buffer.get(offset+1);
        byte res;
        String s=new String();
        for (int i=offset+2;i<offset+2+l;i++)
        {
            res=buffer.get(i);
            s+=convertToJava(res);         // convert char if need be
        }
        return s;
    }
  */
    
}
