/*
 * Copyright (c) W. Lenerz 2023.
 * This code is subject to the GPL, version 3 or later.
 */
package ui;

/**
 * Various methods to get configured data from an ini file.
 * 
 * @author and copyright Wolfgang Lenerz
 * @version  1.00 initial version
 */
public class ConfigData
{
    /**
     * Get the configured data for all serial ports.
     * 
     * @param inifile whence to get the data.
     * 
     * @return a string array with the data. Each element in the array may be null or empty.
     */
    public static String [] getSerPortsData(inifile.IniFile inifile)
    {
        int maxNbr = drivers.ser.SerDriver.getMaxSerialPorts();
        String[] configuredNames = new String[maxNbr];
        for (int i=0;i<configuredNames.length;i++)
        {
            configuredNames[i]=inifile.getOptionValue("SER"+(i+1));
        }
        return configuredNames;
    }
    /**
     * Get the configured data for all serial ports.
     * 
     * @param serNames names for serial ports
     * 
     * @param inifile where to get the data.
     * 
     */
    public static void setSerPortsData( String [] serNames,inifile.IniFile inifile)
    {
        for (int n=0;n<serNames.length;n++)
        {
            if (serNames[n]!=null)
                inifile.setOptionValue("SER"+(n+1),serNames[n]);
            else
                inifile.setOptionValue("SER"+(n+1),"");
            inifile.writeIniFileNoError();
        }
    }
    
    /**
     * Sets an option in the inifile and saves that to the file on disk.
     * 
     * @param optionName    name of option (upper cased).
     * @param optionvalue   value of option.
     * @param inifile       inifile containing all options.
     */
    public static void setOptionAndSave(String optionName, String optionvalue,inifile.IniFile inifile)
    {
        inifile.setOptionValue(optionName,optionvalue);
        inifile.writeIniFileNoError();
    }
    
  /**
     * Save new names for a device to the inifile.
     * 
     * @param driveNames    the eight-element array holding the (unresolved) names.
     * @param deviceName    string name of device, 3 letters, uppercased.
     * @param inifile       wither to write the info
    */
    public static void  storeDriveNames(String deviceName,String []driveNames,inifile.IniFile inifile) 
    {
        for (int i=0;i<driveNames.length;i++) 
        {
            inifile.setOptionValue(deviceName+(i+1), driveNames[i]);
        }
        inifile.writeIniFileNoError();       
    }
    
     /**
     * Gets various options from the inifile and sets them. Also shows the state of the options in the GUI Config menu.
     *
     * @param inifile   whence to get the option.
     * @param monitor   monitor with TrapDispatcher object where some options must be set.
     */
    public static void setOptions(inifile.IniFile inifile, smsqmulator.Monitor monitor)
    {        
        smsqmulator.TrapDispatcher td;
        if (monitor!=null && (td = monitor.getTrapDispatcher())!=null)
        {
            td.setNamesForDrives(smsqmulator.Types.WINDriver,drivers.ResolvePath.resolvePaths(getNames("WIN",inifile),true),true); 
            td.setNamesForDrives(smsqmulator.Types.SFADriver,drivers.ResolvePath.resolvePaths(getNames("SFA",inifile),true),true); 
            td.setNamesForDrives(smsqmulator.Types.NFADriver,drivers.ResolvePath.resolvePaths(getNames("NFA",inifile),true),true); 
            td.setNamesForDrives(smsqmulator.Types.MEMDriver,drivers.ResolvePath.resolvePaths(getNames("MEM",inifile),true),true); 

            // set the usage names of inbuilt devices (no need to set name for flp!)
            String device=inifile.getOptionValue("NFA_USE");
            if (device!=null && !device.isEmpty())
                td.setUsageForDrive(smsqmulator.Types.NFADriver, smsqmulator.Helper.convertUsageName(device));  
            device=inifile.getOptionValue("SFA_USE");
            if (device!=null && !device.isEmpty())
                td.setUsageForDrive(smsqmulator.Types.SFADriver, smsqmulator.Helper.convertUsageName(device));  
            device=inifile.getOptionValue("WIN_USE");
            if (device!=null && !device.isEmpty())
                td.setUsageForDrive(smsqmulator.Types.WINDriver, smsqmulator.Helper.convertUsageName(device));
            device=inifile.getOptionValue("MEM_USE");
            if (device!=null && !device.isEmpty())
                td.setUsageForDrive(smsqmulator.Types.MEMDriver, smsqmulator.Helper.convertUsageName(device));

            td.setFilenameChange(smsqmulator.Types.NFADriver,inifile.getOptionAsInt("NFA-FILENAME-CHANGE", 0)); 
            td.setFilenameChange(smsqmulator.Types.SFADriver,inifile.getOptionAsInt("SFA-FILENAME-CHANGE", 0)); 
            td.setFloppyNames(drivers.FloppyDriver.getNativeNames());
        }   
    }
    
    /**
     * Get names for drives.
     * 
     * @param rootName
     * @param inifile
     * @return 
     */
    private static String[]getNames(String rootName,inifile.IniFile inifile)
    {
        String[]result =new String[8];
        for (int i = 1;i<9;i++)
            result[i-1]=inifile.getOptionValue(rootName+i);
        return result;
    }
}
